#!/usr/bin/env python
############################################################################
#  imagecomp.py
#
#  Copyright 2017 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities for creating image composites
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 04/10/2017
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import glob
import os
import os.path
import numpy
import shutil

import osgeo.gdal
import rios.rat

import rsgislib
import rsgislib.imagecalc
import rsgislib.imagecalc.calcindices
import rsgislib.imageutils
import rsgislib.rastergis


def create_max_ndvi_composite(
    in_imgs_pattern,
    r_band,
    n_band,
    out_ref_img,
    out_comp_img,
    tmp_dir="tmp",
    gdalformat="KEA",
    datatype=None,
    calc_stats=True,
):
    """
    Create an image composite from multiple input images where the pixel brought through into the composite is the one with
    the maximum NDVI.

    :param in_imgs_pattern: is a pattern (ready for glob.glob(inImgsPattern)) so needs an '*' within the pattern to find the input image files.
    :param r_band: is the image band within the input images (same for all) for the red band - note band indexing starts at 1.
    :param n_band: is the image band within the input images (same for all) for the nir band - note band indexing starts at 1.
    :param out_ref_img: is the output reference image which details which input image is forming the output image pixel value (Note. this file will always be a KEA file as RAT is used).
    :param out_comp_img: is the output composite image for which gdalformat and dataType define the format and data type.
    :param tmp_dir: is a temp path for intemediate files, if this path doesn't exist is will be created and deleted at runtime.
    :param gdalformat: is the output file format of the outCompImg, any GDAL compatable format is OK (Defaut is KEA).
    :param datatype: is the data type of the output image (outCompImg). If None is provided then the data type of the first input image will be used (Default None).
    :param calc_stats: calculate image statics and pyramids (Default=True)

    """
    import rsgislib.tools.utils

    uidStr = rsgislib.tools.utils.uid_generator()

    # Get List of input images:
    inImages = glob.glob(in_imgs_pattern)

    if len(inImages) > 1:
        tmpPresent = True
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
            tmpPresent = False

        refLayersPath = os.path.join(tmp_dir, "RefLyrs_" + uidStr)

        refImgTmpPresent = True
        if not os.path.exists(refLayersPath):
            os.makedirs(refLayersPath)
            refImgTmpPresent = False

        if datatype is None:
            datatype = rsgislib.imageutils.get_rsgislib_datatype_from_img(inImages[0])

        numInLyrs = len(inImages)

        numpy.random.seed(5)
        red = numpy.random.randint(1, 255, numInLyrs + 1, int)
        red[0] = 0
        numpy.random.seed(2)
        green = numpy.random.randint(1, 255, numInLyrs + 1, int)
        green[0] = 0
        numpy.random.seed(9)
        blue = numpy.random.randint(1, 255, numInLyrs + 1, int)
        blue[0] = 0
        alpha = numpy.zeros_like(blue)
        alpha[...] = 255
        imgLyrs = numpy.empty(numInLyrs + 1, dtype=numpy.dtype("a255"))

        # Generate Comp Ref layers:
        refLyrsLst = []
        idx = 1
        for img in inImages:
            print("In Image (" + str(idx) + "):\t" + img)
            imgLyrs[idx] = os.path.basename(img)
            baseImgName = os.path.splitext(os.path.basename(img))[0]
            refLyrImg = os.path.join(refLayersPath, baseImgName + "_ndvi.kea")
            rsgislib.imagecalc.calcindices.calc_ndvi(
                img, r_band, n_band, refLyrImg, False
            )
            refLyrsLst.append(refLyrImg)
            idx = idx + 1
        imgLyrs[0] = ""

        # Create REF Image
        rsgislib.imagecalc.get_img_idx_for_stat(
            refLyrsLst, out_ref_img, "KEA", -999, rsgislib.SUMTYPE_MAX
        )
        if calc_stats:
            # Pop Ref Image with stats
            rsgislib.rastergis.pop_rat_img_stats(out_ref_img, True, True, True)

            # Open the clumps dataset as a gdal dataset
            ratDataset = osgeo.gdal.Open(out_ref_img, osgeo.gdal.GA_Update)

            # Write colours to RAT
            rios.rat.writeColumn(ratDataset, "Red", red)
            rios.rat.writeColumn(ratDataset, "Green", green)
            rios.rat.writeColumn(ratDataset, "Blue", blue)
            rios.rat.writeColumn(ratDataset, "Alpha", alpha)
            rios.rat.writeColumn(ratDataset, "Image", imgLyrs)

            ratDataset = None

        # Create Composite Image
        rsgislib.imageutils.create_ref_img_composite_img(
            inImages, out_comp_img, out_ref_img, gdalformat, datatype, 0.0
        )

        if calc_stats:
            # Calc Stats
            rsgislib.imageutils.pop_img_stats(
                out_comp_img, use_no_data=True, no_data_val=0, calc_pyramids=True
            )

        if not refImgTmpPresent:
            shutil.rmtree(refLayersPath, ignore_errors=True)

        if not tmpPresent:
            shutil.rmtree(tmp_dir, ignore_errors=True)
    elif len(inImages) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        shutil.copy(inImages[0], out_comp_img)
    else:
        raise rsgislib.RSGISPyException(
            "There were no input images for " + in_imgs_pattern
        )


def create_max_ndvi_ndwi_composite_landsat(
    inImgsPattern,
    outRefImg,
    outCompImg,
    outMskImg,
    tmpPath="tmp",
    gdalformat="KEA",
    dataType=None,
    calcStats=True,
    use_mode=True,
):
    """
    Create an image composite from multiple input images where the pixel brought through into the composite is the one with
    the maximum NDVI over land and NDWI over water. A mask of land and water regions is also produced. Landsat 8 images are
    identified as image file names are expected to start 'LS8'. If composite includes a mix of LS8 and LS7 images then the
    LS8 images are submitted to match the images bands of LS7 (i.e., coastal band removed).

    :param inImgsPattern: is a pattern (ready for glob.glob(inImgsPattern)) so needs an '*' within the pattern to find the input image files.
    :param outRefImg: is the output reference image which details which input image is forming the output image pixel value (Note. this file will always be a KEA file as RAT is used).
    :param outCompImg: is the output composite image for which gdalformat and dataType define the format and data type.
    :param outMskImg: is the output mask image for regions of water and land where ndvi vs ndwi are used (0=nodata, 1=land, 2=water)
    :param tmpPath: is a temp path for intemediate files, if this path doesn't exist is will be created and deleted at runtime.
    :param gdalformat: is the output file format of the outCompImg, any GDAL compatable format is OK (Defaut is KEA).
    :param dataType: is the data type of the output image (outCompImg). If None is provided then the data type of the first input image will be used (Default None).
    :param calcStats: calculate image statics and pyramids (Default=True)
    :param use_mode: True: the land/water masks are combined using the mode and False: the land water masks are combined using median.

    """
    import rsgislib.tools.utils

    uidStr = rsgislib.tools.utils.uid_generator()

    # Get List of input images:
    inImages = glob.glob(inImgsPattern)

    if len(inImages) > 1:
        tmpPresent = True
        if not os.path.exists(tmpPath):
            os.makedirs(tmpPath)
            tmpPresent = False

        refLayersPath = os.path.join(tmpPath, "RefLyrs_" + uidStr)

        refImgTmpPresent = True
        if not os.path.exists(refLayersPath):
            os.makedirs(refLayersPath)
            refImgTmpPresent = False

        if dataType is None:
            dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(inImages[0])

        numInLyrs = len(inImages)

        numpy.random.seed(5)
        red = numpy.random.randint(1, 255, numInLyrs + 1, int)
        red[0] = 0
        numpy.random.seed(2)
        green = numpy.random.randint(1, 255, numInLyrs + 1, int)
        green[0] = 0
        numpy.random.seed(9)
        blue = numpy.random.randint(1, 255, numInLyrs + 1, int)
        blue[0] = 0
        alpha = numpy.zeros_like(blue)
        alpha[...] = 255
        imgLyrs = numpy.empty(numInLyrs + 1, dtype=numpy.dtype("a255"))

        # Generate Comp Ref layers:
        landWaterMskBandDefns = []
        mskImgs = []
        idx = 1
        onlyLS8 = True
        incLS8 = False
        for img in inImages:
            print("In Image (" + str(idx) + "):\t" + img)
            imgLyrs[idx] = os.path.basename(img)
            baseImgName = os.path.splitext(os.path.basename(img))[0]
            refLyrNDVIImg = os.path.join(refLayersPath, baseImgName + "_ndvi.kea")
            refLyrNDWIImg = os.path.join(refLayersPath, baseImgName + "_ndwi.kea")
            if "LS8" in img:
                rBand = 4
                nBand = 5
                sBand = 6
                incLS8 = True
            else:
                rBand = 3
                nBand = 4
                sBand = 5
                onlyLS8 = False

            rsgislib.imagecalc.calcindices.calc_ndvi(
                img, rBand, nBand, refLyrNDVIImg, False
            )
            rsgislib.imagecalc.calcindices.calc_ndwi(
                img, nBand, sBand, refLyrNDWIImg, False
            )

            refLyrMskImg = os.path.join(
                refLayersPath, baseImgName + "_waterLandMsk.kea"
            )
            bandDefns = []
            bandDefns.append(rsgislib.imagecalc.BandDefn("ndvi", refLyrNDVIImg, 1))
            bandDefns.append(rsgislib.imagecalc.BandDefn("ndwi", refLyrNDWIImg, 1))
            rsgislib.imagecalc.band_math(
                refLyrMskImg,
                "ndvi<-1?0:ndvi>0.3?1:ndwi>0.01?2:1",
                "KEA",
                rsgislib.TYPE_8UINT,
                bandDefns,
            )
            mskImgs.append(refLyrMskImg)
            idx = idx + 1
        imgLyrs[0] = ""

        refLyrMskStackImg = os.path.join(
            refLayersPath, uidStr + "_waterLandMskStack.kea"
        )
        rsgislib.imageutils.stack_img_bands(
            mskImgs, None, refLyrMskStackImg, -1, -1, "KEA", rsgislib.TYPE_8UINT
        )
        if use_mode:
            rsgislib.imagecalc.imagePixelColumnSummary(
                refLyrMskStackImg,
                outMskImg,
                rsgislib.imagecalc.StatsSummary(calcMode=True),
                "KEA",
                rsgislib.TYPE_8UINT,
                0,
                True,
            )
        else:
            rsgislib.imagecalc.imagePixelColumnSummary(
                refLyrMskStackImg,
                outMskImg,
                rsgislib.imagecalc.StatsSummary(calcMedian=True),
                "KEA",
                rsgislib.TYPE_8UINT,
                0,
                True,
            )
        rsgislib.rastergis.pop_rat_img_stats(outMskImg, True, True, True)

        idx = 1
        refLyrsLst = []
        for img in inImages:
            print("In Image (" + str(idx) + "):\t" + img)
            baseImgName = os.path.splitext(os.path.basename(img))[0]
            refLyrNDVIImg = os.path.join(refLayersPath, baseImgName + "_ndvi.kea")
            refLyrNDWIImg = os.path.join(refLayersPath, baseImgName + "_ndwi.kea")
            refLyrLclMskImg = os.path.join(
                refLayersPath, baseImgName + "_waterLandMsk.kea"
            )

            refLyrImg = os.path.join(refLayersPath, baseImgName + "_refHybrid.kea")
            bandDefns = []
            bandDefns.append(rsgislib.imagecalc.BandDefn("lmsk", refLyrLclMskImg, 1))
            bandDefns.append(rsgislib.imagecalc.BandDefn("omsk", outMskImg, 1))
            bandDefns.append(rsgislib.imagecalc.BandDefn("ndvi", refLyrNDVIImg, 1))
            bandDefns.append(rsgislib.imagecalc.BandDefn("ndwi", refLyrNDWIImg, 1))
            rsgislib.imagecalc.band_math(
                refLyrImg,
                "lmsk==0?-999:omsk==1?ndvi:omsk==2?ndwi:-999",
                "KEA",
                rsgislib.TYPE_32FLOAT,
                bandDefns,
            )
            refLyrsLst.append(refLyrImg)
            idx = idx + 1

        if incLS8 and (not onlyLS8):
            inImagesTmp = []
            for img in inImages:
                if "LS8" in img:
                    baseImgName = os.path.splitext(os.path.basename(img))[0]
                    reflSubImg = os.path.join(tmpPath, baseImgName + "_bandsub.kea")
                    rsgislib.imageutils.select_img_bands(
                        img, reflSubImg, "KEA", dataType, [2, 3, 4, 5, 6, 7]
                    )
                    inImagesTmp.append(reflSubImg)
                else:
                    inImagesTmp.append(img)
            inImages = inImagesTmp

        # Create REF Image
        rsgislib.imagecalc.get_img_idx_for_stat(
            refLyrsLst, outRefImg, "KEA", -999, rsgislib.SUMTYPE_MAX
        )
        if calcStats:
            # Pop Ref Image with stats
            rsgislib.rastergis.pop_rat_img_stats(outRefImg, True, True, True)

            # Open the clumps dataset as a gdal dataset
            ratDataset = osgeo.gdal.Open(outRefImg, osgeo.gdal.GA_Update)
            # Write colours to RAT
            rios.rat.writeColumn(ratDataset, "Red", red)
            rios.rat.writeColumn(ratDataset, "Green", green)
            rios.rat.writeColumn(ratDataset, "Blue", blue)
            rios.rat.writeColumn(ratDataset, "Alpha", alpha)
            rios.rat.writeColumn(ratDataset, "Image", imgLyrs)
            ratDataset = None

        # Create Composite Image
        rsgislib.imageutils.create_ref_img_composite_img(
            inImages, outCompImg, outRefImg, gdalformat, dataType, 0.0
        )

        if calcStats:
            # Calc Stats
            rsgislib.imageutils.pop_img_stats(
                outCompImg, use_no_data=True, no_data_val=0, calc_pyramids=True
            )

        if not refImgTmpPresent:
            shutil.rmtree(refLayersPath, ignore_errors=True)

        if not tmpPresent:
            shutil.rmtree(tmpPath, ignore_errors=True)

    elif len(inImages) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        shutil.copy(inImages[0], outCompImg)
    else:
        raise rsgislib.RSGISPyException(
            "There were no input images for " + inImgsPattern
        )


def check_build_ls8_vrts(input_imgs, output_dir):
    """
    A function which checks for Landsat 8 (LS8) images in the input list and creates
    a band subset (i.e., removes the coastal (1) band). This function should be used
    ahead of create_max_ndvi_ndwi_composite when combining LS8 with earlier landsat
    (i.e., LS5 and LS7) data. Note, files who's file name contain 'LS8' are considered
    to be Landsat 8 images.

    :param input_imgs: input list of image file paths.
    :param output_dir: output directory for the VRT images (Note. an absolute path to
                   the input image is used when building the VRT.)
    """
    import rsgislib.tools.filetools

    img_bands = [2, 3, 4, 5, 6, 7]
    out_imgs = []
    for img in input_imgs:
        if "LS8" in os.path.basename(img):
            abs_img = os.path.abspath(img)
            basename = rsgislib.tools.filetools.get_file_basename(img)
            out_vrt_img = os.path.join(output_dir, f"{basename}_bsub.vrt")
            rsgislib.imageutils.create_vrt_band_subset(abs_img, img_bands, out_vrt_img)
            out_imgs.append(out_vrt_img)
        else:
            out_imgs.append(img)
    return out_imgs


def create_max_ndvi_ndwi_composite(
    refImg,
    inImages,
    rBand,
    nBand,
    sBand,
    outRefImg,
    outCompImg,
    outMskImg,
    tmpPath="tmp",
    gdalformat="KEA",
    dataType=None,
    calcStats=True,
    reprojmethod="cubic",
    use_mode=True,
):
    """
    Create an image composite from multiple input images where the pixel brought through into the composite is the one with
    the maximum NDVI over land and NDWI over water. A mask of land and water regions is also produced. The reference image is
    used to define the spatial extent of the output images and spatial projection.

    :param refImg: is a reference image with any number of bands and data type which is used to define the output image extent and projection.
    :param inImages: a list of all the input images being used to build the composite.
    :param rBand: is an integer specifying the red band in the input images (starts at 1), used in the NDVI index.
    :param nBand: is an integer specifying the NIR band in the input images (starts at 1), used in the NDVI and NDWI index.
    :param sBand: is an integer specifying the SWIR band in the input images (starts at 1), used in the NDVI and NDWI index.
    :param outRefImg: is the output reference image which details which input image is forming the output image pixel value (Note. this file will always be a KEA file as RAT is used).
    :param outCompImg: is the output composite image for which gdalformat and dataType define the format and data type.
    :param outMskImg: is the output mask image for regions of water and land where ndvi vs ndwi are used (0=nodata, 1=land, 2=water)
    :param tmpPath: is a temp path for intemediate files, if this path doesn't exist is will be created and deleted at runtime.
    :param gdalformat: is the output file format of the outCompImg, any GDAL compatable format is OK (Defaut is KEA).
    :param dataType: is the data type of the output image (outCompImg). If None is provided then the data type of the first input image will be used (Default None).
    :param calcStats: calculate image statics and pyramids (Default=True)
    :param reprojmethod: specifies the interpolation method used to reproject the input images which are in a different projection and/or pixel size as the reference image (default: cubic).
    :param use_mode: True: the land/water masks are combined using the mode and False: the land water masks are combined using median.

    """
    import rsgislib.tools.utils

    uidStr = rsgislib.tools.utils.uid_generator()

    if len(inImages) > 1:
        init_in_images = inImages
        inImagesOverlap = []
        inImages2ReProj = []
        inImagesSub2Ref = []
        nBands = 0
        first = True
        for img in init_in_images:
            if first:
                nBands = rsgislib.imageutils.get_img_band_count(img)
                first = False
            else:
                cBands = rsgislib.imageutils.get_img_band_count(img)
                if cBands != nBands:
                    raise rsgislib.RSGISPyException(
                        "The number of image bands is not consistent (Bands: {0} and {1})".format(
                            nBands, cBands
                        )
                    )

            sameProj = rsgislib.imageutils.do_gdal_layers_have_same_proj(refImg, img)
            if rsgislib.imageutils.doImagesOverlap(refImg, img):
                if sameProj:
                    if rsgislib.imageutils.doImageResMatch(refImg, img):
                        inImagesOverlap.append(img)
                    else:
                        inImages2ReProj.append(img)
                else:
                    inImages2ReProj.append(img)

        nImgs = len(inImagesOverlap) + len(inImages2ReProj)
        print("There are {} images with overlap to create the composite.".format(nImgs))
        if nImgs > 1:
            if dataType is None:
                if len(inImagesOverlap) > 0:
                    dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(
                        inImagesOverlap[0]
                    )
                else:
                    dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(
                        inImages2ReProj[0]
                    )

            tmpPresent = True
            if not os.path.exists(tmpPath):
                os.makedirs(tmpPath)
                tmpPresent = False

            subsetImgTmpPresent = True
            if len(inImagesOverlap) > 0:
                print("Subset images to reference extent.")
                subsetLayersPath = os.path.join(tmpPath, "Subset_" + uidStr)
                if not os.path.exists(subsetLayersPath):
                    os.makedirs(subsetLayersPath)
                    subsetImgTmpPresent = False

                for img in inImagesOverlap:
                    basename = os.path.splitext(os.path.basename(img))[0]
                    outImg = os.path.join(subsetLayersPath, basename + "_subset.kea")
                    rsgislib.imageutils.create_copy_img(
                        refImg, outImg, nBands, 0, "KEA", dataType
                    )
                    rsgislib.imageutils.includeImagesIndImgIntersect(outImg, [img])
                    inImagesSub2Ref.append(outImg)

            reprojImgTmpPresent = True
            if len(inImages2ReProj) > 0:
                print("Reproject images to reference extent.")
                reprojLayersPath = os.path.join(tmpPath, "Reproj_" + uidStr)
                if not os.path.exists(reprojLayersPath):
                    os.makedirs(reprojLayersPath)
                    reprojImgTmpPresent = False

                for img in inImages2ReProj:
                    basename = os.path.splitext(os.path.basename(img))[0]
                    outImg = os.path.join(reprojLayersPath, basename + "_reproj.kea")
                    rsgislib.imageutils.resampleImage2Match(
                        refImg, img, outImg, "KEA", reprojmethod, dataType
                    )
                    inImagesSub2Ref.append(outImg)

            refLayersPath = os.path.join(tmpPath, "RefLyrs_" + uidStr)
            refImgTmpPresent = True
            if not os.path.exists(refLayersPath):
                os.makedirs(refLayersPath)
                refImgTmpPresent = False

            numInLyrs = len(inImagesSub2Ref)

            numpy.random.seed(5)
            red = numpy.random.randint(1, 255, numInLyrs + 1, int)
            red[0] = 0
            numpy.random.seed(2)
            green = numpy.random.randint(1, 255, numInLyrs + 1, int)
            green[0] = 0
            numpy.random.seed(9)
            blue = numpy.random.randint(1, 255, numInLyrs + 1, int)
            blue[0] = 0
            alpha = numpy.zeros_like(blue)
            alpha[...] = 255
            imgLyrs = numpy.empty(numInLyrs + 1, dtype=numpy.dtype("a255"))

            # Generate Comp Ref layers:
            landWaterMskBandDefns = []
            mskImgs = []
            idx = 1
            for img in inImagesSub2Ref:
                print("In Image (" + str(idx) + "):\t" + img)
                imgLyrs[idx] = os.path.basename(img)
                baseImgName = os.path.splitext(os.path.basename(img))[0]
                refLyrNDVIImg = os.path.join(refLayersPath, baseImgName + "_ndvi.kea")
                refLyrNDWIImg = os.path.join(refLayersPath, baseImgName + "_ndwi.kea")
                rsgislib.imagecalc.calcindices.calc_ndvi(
                    img, rBand, nBand, refLyrNDVIImg, False
                )
                rsgislib.imagecalc.calcindices.calc_ndwi(
                    img, nBand, sBand, refLyrNDWIImg, False
                )

                refLyrMskImg = os.path.join(
                    refLayersPath, baseImgName + "_waterLandMsk.kea"
                )
                bandDefns = []
                bandDefns.append(rsgislib.imagecalc.BandDefn("ndvi", refLyrNDVIImg, 1))
                bandDefns.append(rsgislib.imagecalc.BandDefn("ndwi", refLyrNDWIImg, 1))
                rsgislib.imagecalc.band_math(
                    refLyrMskImg,
                    "ndvi<-1?0:ndvi>0.3?1:ndwi>0.01?2:1",
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    bandDefns,
                )
                mskImgs.append(refLyrMskImg)
                idx = idx + 1
            imgLyrs[0] = ""

            refLyrMskStackImg = os.path.join(
                refLayersPath, uidStr + "_waterLandMskStack.kea"
            )
            rsgislib.imageutils.stack_img_bands(
                mskImgs, None, refLyrMskStackImg, -1, -1, "KEA", rsgislib.TYPE_8UINT
            )
            if use_mode:
                rsgislib.imagecalc.imagePixelColumnSummary(
                    refLyrMskStackImg,
                    outMskImg,
                    rsgislib.imagecalc.StatsSummary(calcMode=True),
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    0,
                    True,
                )
            else:
                rsgislib.imagecalc.imagePixelColumnSummary(
                    refLyrMskStackImg,
                    outMskImg,
                    rsgislib.imagecalc.StatsSummary(calcMedian=True),
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    0,
                    True,
                )
            rsgislib.rastergis.pop_rat_img_stats(outMskImg, True, True, True)

            idx = 1
            refLyrsLst = []
            for img in inImagesSub2Ref:
                print("In Image (" + str(idx) + "):\t" + img)
                baseImgName = os.path.splitext(os.path.basename(img))[0]
                refLyrNDVIImg = os.path.join(refLayersPath, baseImgName + "_ndvi.kea")
                refLyrNDWIImg = os.path.join(refLayersPath, baseImgName + "_ndwi.kea")
                refLyrLclMskImg = os.path.join(
                    refLayersPath, baseImgName + "_waterLandMsk.kea"
                )

                refLyrImg = os.path.join(refLayersPath, baseImgName + "_refHybrid.kea")
                bandDefns = []
                bandDefns.append(
                    rsgislib.imagecalc.BandDefn("lmsk", refLyrLclMskImg, 1)
                )
                bandDefns.append(rsgislib.imagecalc.BandDefn("omsk", outMskImg, 1))
                bandDefns.append(rsgislib.imagecalc.BandDefn("ndvi", refLyrNDVIImg, 1))
                bandDefns.append(rsgislib.imagecalc.BandDefn("ndwi", refLyrNDWIImg, 1))
                rsgislib.imagecalc.band_math(
                    refLyrImg,
                    "lmsk==0?-999:omsk==1?ndvi:omsk==2?ndwi:-999",
                    "KEA",
                    rsgislib.TYPE_32FLOAT,
                    bandDefns,
                )
                refLyrsLst.append(refLyrImg)
                idx = idx + 1

            # Create REF Image
            rsgislib.imagecalc.get_img_idx_for_stat(
                refLyrsLst, outRefImg, "KEA", -999, rsgislib.SUMTYPE_MAX
            )
            if calcStats:
                # Pop Ref Image with stats
                rsgislib.rastergis.pop_rat_img_stats(outRefImg, True, True, True)

                # Open the clumps dataset as a gdal dataset
                ratDataset = osgeo.gdal.Open(outRefImg, osgeo.gdal.GA_Update)
                # Write colours to RAT
                rios.rat.writeColumn(ratDataset, "Red", red)
                rios.rat.writeColumn(ratDataset, "Green", green)
                rios.rat.writeColumn(ratDataset, "Blue", blue)
                rios.rat.writeColumn(ratDataset, "Alpha", alpha)
                rios.rat.writeColumn(ratDataset, "Image", imgLyrs)
                ratDataset = None

            # Create Composite Image
            rsgislib.imageutils.create_ref_img_composite_img(
                inImagesSub2Ref, outCompImg, outRefImg, gdalformat, dataType, 0.0
            )

            if calcStats:
                # Calc Stats
                rsgislib.imageutils.pop_img_stats(
                    outCompImg, use_no_data=True, no_data_val=0, calc_pyramids=True
                )

            if not refImgTmpPresent:
                shutil.rmtree(refLayersPath, ignore_errors=True)

            if not subsetImgTmpPresent:
                shutil.rmtree(subsetLayersPath, ignore_errors=True)

            if not reprojImgTmpPresent:
                shutil.rmtree(reprojLayersPath, ignore_errors=True)

            if not tmpPresent:
                shutil.rmtree(tmpPath, ignore_errors=True)
        else:
            raise rsgislib.RSGISPyException(
                "There were only {0} images which intersect with the reference image.".format(
                    nImgs
                )
            )

    elif len(inImages) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        nBands = rsgislib.imageutils.get_img_band_count(inImages[0])
        if dataType is None:
            dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(inImages[0])
        rsgislib.imageutils.create_copy_img(
            refImg, outCompImg, nBands, 0, gdalformat, dataType
        )
        rsgislib.imageutils.includeImagesIndImgIntersect(outCompImg, [inImages[0]])
    else:
        raise rsgislib.RSGISPyException("List of input images was empty ")


def create_max_scaled_ndvi_ndwi_composite(
    refImg,
    inImages,
    gBand,
    rBand,
    nBand,
    sBand,
    outRefImg,
    outCompImg,
    outMskImg,
    tmpPath="tmp",
    gdalformat="KEA",
    dataType=None,
    calcStats=True,
    reprojmethod="cubic",
    use_mode=True,
):
    """
    Create an image composite from multiple input images where the pixel brought through into the composite is the one with
    the maximum NDVI over land and NDWI over water. A mask of land and water regions is also produced. The reference image is
    used to define the spatial extent of the output images and spatial projection.

    :param refImg: is a reference image with any number of bands and data type which is used to define the output image extent and projection.
    :param inImages: a list of all the input images being used to build the composite.
    :param gBand: is an integer specifying the green band in the input images (starts at 1), used in the GNDWI index.
    :param rBand: is an integer specifying the red band in the input images (starts at 1), used in the NDVI index.
    :param nBand: is an integer specifying the NIR band in the input images (starts at 1), used in the NDVI, GNDWI and NDWI index.
    :param sBand: is an integer specifying the SWIR band in the input images (starts at 1), used in the NDVI and NDWI index.
    :param outRefImg: is the output reference image which details which input image is forming the output image pixel value (Note. this file will always be a KEA file as RAT is used).
    :param outCompImg: is the output composite image for which gdalformat and dataType define the format and data type.
    :param outMskImg: is the output mask image for regions of water and land where ndvi vs ndwi are used (0=nodata, 1=land, 2=water)
    :param tmpPath: is a temp path for intemediate files, if this path doesn't exist is will be created and deleted at runtime.
    :param gdalformat: is the output file format of the outCompImg, any GDAL compatable format is OK (Defaut is KEA).
    :param dataType: is the data type of the output image (outCompImg). If None is provided then the data type of the first input image will be used (Default None).
    :param calcStats: calculate image statics and pyramids (Default=True)
    :param reprojmethod: specifies the interpolation method used to reproject the input images which are in a different projection and/or pixel size as the reference image (default: cubic).
    :param use_mode: True: the land/water masks are combined using the mode and False: the land water masks are combined using median.

    """
    import rsgislib.tools.utils

    uidStr = rsgislib.tools.utils.uid_generator()

    if len(inImages) > 1:
        init_in_images = inImages
        inImagesOverlap = []
        inImages2ReProj = []
        inImagesSub2Ref = []
        nBands = 0
        first = True
        for img in init_in_images:
            if first:
                nBands = rsgislib.imageutils.get_img_band_count(img)
                first = False
            else:
                cBands = rsgislib.imageutils.get_img_band_count(img)
                if cBands != nBands:
                    raise rsgislib.RSGISPyException(
                        "The number of image bands is not consistent (Bands: {0} and {1})".format(
                            nBands, cBands
                        )
                    )

            sameProj = rsgislib.imageutils.do_gdal_layers_have_same_proj(refImg, img)
            if rsgislib.imageutils.doImagesOverlap(refImg, img):
                if sameProj:
                    if rsgislib.imageutils.doImageResMatch(refImg, img):
                        inImagesOverlap.append(img)
                    else:
                        inImages2ReProj.append(img)
                else:
                    inImages2ReProj.append(img)

        nImgs = len(inImagesOverlap) + len(inImages2ReProj)
        print("There are {} images with overlap to create the composite.".format(nImgs))
        if nImgs > 1:
            if dataType is None:
                if len(inImagesOverlap) > 0:
                    dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(
                        inImagesOverlap[0]
                    )
                else:
                    dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(
                        inImages2ReProj[0]
                    )

            tmpPresent = True
            if not os.path.exists(tmpPath):
                os.makedirs(tmpPath)
                tmpPresent = False

            subsetImgTmpPresent = True
            if len(inImagesOverlap) > 0:
                print("Subset images to reference extent.")
                subsetLayersPath = os.path.join(tmpPath, "Subset_" + uidStr)
                if not os.path.exists(subsetLayersPath):
                    os.makedirs(subsetLayersPath)
                    subsetImgTmpPresent = False

                for img in inImagesOverlap:
                    basename = os.path.splitext(os.path.basename(img))[0]
                    outImg = os.path.join(subsetLayersPath, basename + "_subset.kea")
                    rsgislib.imageutils.create_copy_img(
                        refImg, outImg, nBands, 0, "KEA", dataType
                    )
                    rsgislib.imageutils.includeImagesIndImgIntersect(outImg, [img])
                    inImagesSub2Ref.append(outImg)

            reprojImgTmpPresent = True
            if len(inImages2ReProj) > 0:
                print("Reproject images to reference extent.")
                reprojLayersPath = os.path.join(tmpPath, "Reproj_" + uidStr)
                if not os.path.exists(reprojLayersPath):
                    os.makedirs(reprojLayersPath)
                    reprojImgTmpPresent = False

                for img in inImages2ReProj:
                    basename = os.path.splitext(os.path.basename(img))[0]
                    outImg = os.path.join(reprojLayersPath, basename + "_reproj.kea")
                    rsgislib.imageutils.resampleImage2Match(
                        refImg, img, outImg, "KEA", reprojmethod, dataType
                    )
                    inImagesSub2Ref.append(outImg)

            refLayersPath = os.path.join(tmpPath, "RefLyrs_" + uidStr)
            refImgTmpPresent = True
            if not os.path.exists(refLayersPath):
                os.makedirs(refLayersPath)
                refImgTmpPresent = False

            numInLyrs = len(inImagesSub2Ref)

            numpy.random.seed(5)
            red = numpy.random.randint(1, 255, numInLyrs + 1, int)
            red[0] = 0
            numpy.random.seed(2)
            green = numpy.random.randint(1, 255, numInLyrs + 1, int)
            green[0] = 0
            numpy.random.seed(9)
            blue = numpy.random.randint(1, 255, numInLyrs + 1, int)
            blue[0] = 0
            alpha = numpy.zeros_like(blue)
            alpha[...] = 255
            imgLyrs = numpy.empty(numInLyrs + 1, dtype=numpy.dtype("a255"))

            # Generate Comp Ref layers:
            landWaterMskBandDefns = []
            mskImgs = []
            idx = 1
            for img in inImagesSub2Ref:
                print("In Image (" + str(idx) + "):\t" + img)
                imgLyrs[idx] = os.path.basename(img)
                baseImgName = os.path.splitext(os.path.basename(img))[0]
                refLyrNDVIImg = os.path.join(refLayersPath, baseImgName + "_ndvi.kea")
                refLyrNDWIImg = os.path.join(refLayersPath, baseImgName + "_ndwi.kea")
                ###refLyrGNDWIImg = os.path.join(refLayersPath, baseImgName+'_gndwi.kea')
                ###refLyrGMNDWIImg = os.path.join(refLayersPath, baseImgName+'_gmndwi.kea')
                rsgislib.imagecalc.calcindices.calc_ndvi(
                    img, rBand, nBand, refLyrNDVIImg, False
                )
                rsgislib.imagecalc.calcindices.calc_ndwi(
                    img, nBand, sBand, refLyrNDWIImg, False
                )
                ###rsgislib.imagecalc.calcindices.calc_gndwi(img, gBand, nBand, refLyrGNDWIImg, False)
                ###rsgislib.imagecalc.calcindices.calc_gmndwi(img, gBand, sBand, refLyrGMNDWIImg, False)

                refLyrMskImg = os.path.join(
                    refLayersPath, baseImgName + "_waterLandMsk.kea"
                )
                bandDefns = []
                bandDefns.append(rsgislib.imagecalc.BandDefn("ndvi", refLyrNDVIImg, 1))
                bandDefns.append(rsgislib.imagecalc.BandDefn("ndwi", refLyrNDWIImg, 1))
                ###bandDefns.append(rsgislib.imagecalc.BandDefn('gndwi', refLyrGNDWIImg, 1))
                ###bandDefns.append(rsgislib.imagecalc.BandDefn('gmndwi', refLyrGMNDWIImg, 1))
                ###rsgislib.imagecalc.band_math(refLyrMskImg, '(ndvi<-999)?0:(ndvi>0.3)?1:(ndwi>0.01)?2:(gndwi>0.01)?2:(gmndwi>0.01)?2:1', 'KEA', rsgislib.TYPE_8UINT, bandDefns)
                rsgislib.imagecalc.band_math(
                    refLyrMskImg,
                    "(ndvi==-999)?0:(ndvi>0.3)?1:(ndwi>0.01)?2:1",
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    bandDefns,
                )
                mskImgs.append(refLyrMskImg)
                idx = idx + 1
            imgLyrs[0] = ""

            print("Creating land/Water mask.")
            refLyrMskStackImg = os.path.join(
                refLayersPath, uidStr + "_waterLandMskStack.kea"
            )
            rsgislib.imageutils.stack_img_bands(
                mskImgs, None, refLyrMskStackImg, -1, -1, "KEA", rsgislib.TYPE_8UINT
            )
            if use_mode:
                rsgislib.imagecalc.imagePixelColumnSummary(
                    refLyrMskStackImg,
                    outMskImg,
                    rsgislib.imagecalc.StatsSummary(calcMode=True),
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    0,
                    True,
                )
            else:
                rsgislib.imagecalc.imagePixelColumnSummary(
                    refLyrMskStackImg,
                    outMskImg,
                    rsgislib.imagecalc.StatsSummary(calcMedian=True),
                    "KEA",
                    rsgislib.TYPE_8UINT,
                    0,
                    True,
                )
            rsgislib.rastergis.pop_rat_img_stats(outMskImg, True, True, True)
            print("Created mask")

            idx = 1
            refLyrsLst = []
            print("create merged NDVI/NDWI layers")
            for img in inImagesSub2Ref:
                print("In Image (" + str(idx) + "):\t" + img)
                baseImgName = os.path.splitext(os.path.basename(img))[0]
                refLyrNDVIImg = os.path.join(
                    refLayersPath, "{}_ndvi.kea".format(baseImgName)
                )
                refLyrNDWIImg = os.path.join(
                    refLayersPath, "{}_ndwi.kea".format(baseImgName)
                )
                refLyrLclMskImg = os.path.join(
                    refLayersPath, "{}_waterLandMsk.kea".format(baseImgName)
                )

                nir_min, nir_max = rsgislib.imagecalc.get_img_band_min_max(
                    img, nBand, True, 0.0
                )

                nir_bright_img = os.path.join(
                    refLayersPath, "{}_nir_brightness.kea".format(baseImgName)
                )
                band_defns = [rsgislib.imagecalc.BandDefn("nir", img, nBand)]
                nir_range = nir_max - nir_min
                exp = "(nir==0)?0.0:(nir-{0})/{1}".format(nir_min, nir_range)
                rsgislib.imagecalc.band_math(
                    nir_bright_img, exp, "KEA", rsgislib.TYPE_32FLOAT, band_defns
                )

                reflyr_ndvi_scaled_img = os.path.join(
                    refLayersPath, "{}_ndvi_scaled.kea".format(baseImgName)
                )
                band_defns = [
                    rsgislib.imagecalc.BandDefn("ndvi", refLyrNDVIImg, 1),
                    rsgislib.imagecalc.BandDefn("bright", nir_bright_img, 1),
                ]
                exp = "(ndvi==-999)?-999:ndvi*bright"
                rsgislib.imagecalc.band_math(
                    reflyr_ndvi_scaled_img,
                    exp,
                    "KEA",
                    rsgislib.TYPE_32FLOAT,
                    band_defns,
                )

                refLyrImg = os.path.join(
                    refLayersPath, "{}_refHybrid.kea".format(baseImgName)
                )
                bandDefns = []
                bandDefns.append(
                    rsgislib.imagecalc.BandDefn("lmsk", refLyrLclMskImg, 1)
                )
                bandDefns.append(rsgislib.imagecalc.BandDefn("omsk", outMskImg, 1))
                bandDefns.append(
                    rsgislib.imagecalc.BandDefn("ndvi", reflyr_ndvi_scaled_img, 1)
                )
                bandDefns.append(rsgislib.imagecalc.BandDefn("ndwi", refLyrNDWIImg, 1))
                rsgislib.imagecalc.band_math(
                    refLyrImg,
                    "lmsk==0?-999:omsk==1?ndvi:omsk==2?ndwi:-999",
                    "KEA",
                    rsgislib.TYPE_32FLOAT,
                    bandDefns,
                )
                refLyrsLst.append(refLyrImg)
                idx = idx + 1

            print("Create reference image..")
            # Create REF Image
            rsgislib.imagecalc.get_img_idx_for_stat(
                refLyrsLst, outRefImg, "KEA", -999, rsgislib.SUMTYPE_MAX
            )
            if calcStats:
                # Pop Ref Image with stats
                rsgislib.rastergis.pop_rat_img_stats(outRefImg, True, True, True)

                # Open the clumps dataset as a gdal dataset
                ratDataset = osgeo.gdal.Open(outRefImg, osgeo.gdal.GA_Update)
                # Write colours to RAT
                rios.rat.writeColumn(ratDataset, "Red", red)
                rios.rat.writeColumn(ratDataset, "Green", green)
                rios.rat.writeColumn(ratDataset, "Blue", blue)
                rios.rat.writeColumn(ratDataset, "Alpha", alpha)
                rios.rat.writeColumn(ratDataset, "Image", imgLyrs)
                ratDataset = None

            # Create Composite Image
            rsgislib.imageutils.create_ref_img_composite_img(
                inImagesSub2Ref, outCompImg, outRefImg, gdalformat, dataType, 0.0
            )

            if calcStats:
                # Calc Stats
                rsgislib.imageutils.pop_img_stats(
                    outCompImg, use_no_data=True, no_data_val=0, calc_pyramids=True
                )

            if not refImgTmpPresent:
                shutil.rmtree(refLayersPath, ignore_errors=True)

            if not subsetImgTmpPresent:
                shutil.rmtree(subsetLayersPath, ignore_errors=True)

            if not reprojImgTmpPresent:
                shutil.rmtree(reprojLayersPath, ignore_errors=True)

            if not tmpPresent:
                shutil.rmtree(tmpPath, ignore_errors=True)
        else:
            raise rsgislib.RSGISPyException(
                "There were only {0} images which intersect with the reference image.".format(
                    nImgs
                )
            )

    elif len(inImages) == 1:
        print("Only 1 Input Image, Just Copying File to output")
        nBands = rsgislib.imageutils.get_img_band_count(inImages[0])
        if dataType is None:
            dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(inImages[0])
        rsgislib.imageutils.create_copy_img(
            refImg, outCompImg, nBands, 0, gdalformat, dataType
        )
        rsgislib.imageutils.includeImagesIndImgIntersect(outCompImg, [inImages[0]])
    else:
        raise rsgislib.RSGISPyException("List of input images was empty ")



#!/usr/bin/env python
"""
The tools file manipulation module contains functions for manipulating and
moving files around.
"""

# Import modules
import glob
import os
import sys
import errno
import tempfile
import shutil
import datetime
import time
from typing import Union, List

import rsgislib


def get_file_basename(
    input_file: str, check_valid: bool = False, n_comps: int = 0, rm_n_exts: int = 0
) -> str:
    """
    Uses os.path module to return file basename (i.e., path and extension removed)

    :param input_file: string for the input file name and path
    :param check_valid: if True then resulting basename will be checked for punctuation
                       characters (other than underscores) and spaces, punctuation
                       will be either removed and spaces changed to an underscore.
                       (Default = False)
    :param n_comps: if > 0 then the resulting basename will be split using underscores
                    and the return based name will be defined using the n_comps
                    components split by under scores.
    :param rm_n_exts: used where an input file has more than one extension
                      (e.g., tar.gz) and only n extensions should be removed.
                      Default: 0 which will removed all extensions calculated
                      based on the number of full-stops (.) within the file name.
                      If a value of 1 was provided for filename.tar.gz then the
                      returns output would be filename.tar.
    :return: basename for file

    """
    import string

    file_base_name = os.path.basename(input_file)
    n_exts = file_base_name.count(".")
    if (rm_n_exts == 0) or (rm_n_exts > n_exts):
        rm_n_exts = n_exts

    basename = file_base_name
    if n_exts > 0:
        for i in range(rm_n_exts):
            basename = os.path.splitext(basename)[0]

    if check_valid:
        basename = basename.replace(" ", "_")
        for punct in string.punctuation:
            if (punct != "_") and (punct != "-"):
                basename = basename.replace(punct, "")
    if n_comps > 0:
        basename_split = basename.split("_")
        if len(basename_split) < n_comps:
            raise rsgislib.RSGISPyException(
                "The number of components specified is more than the number "
                "of components in the basename."
            )
        out_basename = ""
        for i in range(n_comps):
            if i == 0:
                out_basename = basename_split[i]
            else:
                out_basename = out_basename + "_" + basename_split[i]
        basename = out_basename
    return basename


def is_path_valid(file_path_name: str) -> bool:
    """
    This function tests whether a file path is valid in terms of the length of
    each component and the characters used. Should be cross platform.

    This function was adapted from the answer here:
    https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta

    :param file_path_name: the file path to be tested
    :return: True if the passed pathname is a valid pathname for the current OS;

    """
    # Windows-specific error code indicating an invalid pathname.
    ERROR_INVALID_NAME = 123
    # See Also
    # ----------
    # https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-
    #    Official listing of all such codes.

    # If this pathname is either not a string or is but is empty, this pathname
    # is invalid.
    try:
        if not isinstance(file_path_name, str) or not file_path_name:
            return False

        # Strip this pathname's Windows-specific drive specifier (e.g., `C:\`)
        # if any. Since Windows prohibits path components from containing `:`
        # characters, failing to strip this `:`-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, file_path_name = os.path.splitdrive(file_path_name)

        # Directory guaranteed to exist. If the current OS is Windows, this is
        # the drive to which Windows was installed (e.g., the "%HOMEDRIVE%"
        # environment variable); else, the typical root directory.
        root_dirname = (
            os.environ.get("HOMEDRIVE", "C:")
            if sys.platform == "win32"
            else os.path.sep
        )
        assert os.path.isdir(root_dirname)  # ...Murphy and her ironclad Law

        # Append a path separator to this directory if needed.
        root_dirname = root_dirname.rstrip(os.path.sep) + os.path.sep

        # Test whether each path component split from this pathname is valid or
        # not, ignoring non-existent and non-readable path components.
        for pathname_part in file_path_name.split(os.path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code
            # indicates whether this pathname is valid or not. Unless this
            # is the case, this exception implies an ignorable kernel or
            # filesystem complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except ValueError as e:
                return False
            except OSError as exc:
                if hasattr(exc, "winerror"):
                    if exc.winerror == ERROR_INVALID_NAME:
                        return False
                elif exc.errno in {errno.ENAMETOOLONG, errno.ERANGE}:
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError as exc:
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid.
    else:
        return True


def is_path_sibling_creatable(file_path_name: str) -> bool:
    """
    A function which checks whether the file path provided would
    be creatable (i.e., directory exists and you have permission to
    write to it.

    This function was adapted from the answer here:
    https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta

    :param file_path_name: the input file path to be checked.
    :return: True if the current user has sufficient permissions to create **siblings**
             (i.e., arbitrary files in the parent directory) of the passed pathname

    """
    # Parent directory of the passed path. If empty, we substitute the current
    # working directory (CWD) instead.
    dirname = os.path.dirname(file_path_name) or os.getcwd()

    try:
        # For safety, explicitly close and hence delete this temporary file
        # immediately after creating it in the passed path's parent directory.
        with tempfile.TemporaryFile(dir=dirname):
            pass
        return True
    # While the exact type of exception raised by the above function depends on
    # the current version of the Python interpreter, all such types subclass the
    # following exception superclass.
    except EnvironmentError:
        return False


def does_path_exists_or_creatable(file_path_name: str) -> bool:
    """
    A function which checks whether a file path either exists or is
    creatable if it does not exist.

    This function was adapted from the answer here:
    https://stackoverflow.com/questions/9532499/check-whether-a-path-is-valid-in-python-without-creating-a-file-at-the-paths-ta

    :param file_path_name: the input file path to be checked.
    :return: True if the passed pathname is a valid pathname on the current OS _and_
             either currently exists or is hypothetically creatable in a cross-platform
             manner optimized for POSIX-unfriendly filesystems

    """
    try:
        # To prevent "os" module calls from raising undesirable exceptions on
        # invalid path names, is_path_valid() is explicitly called first.
        return is_path_valid(file_path_name) and (
            os.path.exists(file_path_name) or is_path_sibling_creatable(file_path_name)
        )
    # Report failure on non-fatal filesystem complaints (e.g., connection
    # timeouts, permissions issues) implying this path to be inaccessible. All
    # other exceptions are unrelated fatal issues and should not be caught here.
    except OSError:
        return False


def get_dir_name(input_file: str) -> str:
    """
    A function which returns just the name of the directory of the input path
    (file or directory) without the rest of the path.

    :param input_file: string for the input path (file or directory) name and path
    :return: directory name

    """
    input_file = os.path.abspath(input_file)
    if os.path.isfile(input_file):
        dir_path = os.path.dirname(input_file)
    elif os.path.isdir(input_file):
        dir_path = input_file
    else:
        raise rsgislib.RSGISPyException("Input path must be either a file or directory")
    dir_name = os.path.basename(dir_path)
    return dir_name


def split_path_all(input_path: str) -> List[str]:
    """
    A function which splits all the components within a file path
    into a list of components rather than the os.path.split function
    which just splits the last item.

    :param input_path: the input file path.
    :return: a list of the file path components.

    """
    all_parts = []
    while 1:
        parts = os.path.split(input_path)
        if parts[0] == input_path:  # sentinel for absolute paths
            all_parts.insert(0, parts[0])
            break
        elif parts[1] == input_path:  # sentinel for relative paths
            all_parts.insert(0, parts[1])
            break
        else:
            input_path = parts[0]
            all_parts.insert(0, parts[1])
    return all_parts


def delete_file_silent(input_file: str) -> bool:
    """
    A function which can be used in-place of os.remove to delete
    a file but if checks if the file exists and only calls os.remove
    if it does exist but also catches any Exceptions from os.remove
    and just returns a boolean as to whether the input_file has been
    removed.

    :param input_file: input file path for the file which is to be removed.
    :return: boolean (True: File was removed or did not exist. False:
             os.remove through an Exception so assume file was not removed)

    """
    try:
        if os.path.exists(input_file) and os.path.isfile(input_file):
            os.remove(input_file)
        file_removed = True
    except Exception:
        file_removed = False
    return file_removed


def delete_file_with_basename(input_file: str, print_rms=True):
    """
    Function to delete all the files which have a path
    and base name defined in the input_file attribute.

    :param input_file: string for the input file name and path
    :param print_rms: print the files being deleted (Default: True)

    """
    baseName = os.path.splitext(input_file)[0]
    fileList = glob.glob(baseName + str(".*"))
    for file in fileList:
        if print_rms:
            print("Deleting file: " + str(file))
        delete_file_silent(file)


def find_file(dir_path: str, file_search: str) -> str:
    """
    Search for a single file with a path using glob. Therefore, the file
    path returned is a true path. Within the file_search provide the file
    name with '*' as wildcard(s).

    :param dir_path: string for the input directory path
    :param file_search: string with a * wildcard for the file being searched for.
    :return: string with the path to the file

    .. code:: python

        import rsgislib.tools.filetools
        file_path = rsgislib.tools.filetools.find_file("in/dir", "*N15W093*.tif")

    """
    files = glob.glob(os.path.join(dir_path, file_search))
    if len(files) != 1:
        raise rsgislib.RSGISPyException(
            "Could not find a single file ("
            + file_search
            + "); found "
            + str(len(files))
            + " files."
        )
    return files[0]


def find_file_none(dir_path: str, file_search: str) -> Union[None, str]:
    """
    Search for a single file with a path using glob. Therefore, the file
    path returned is a true path. Within the file_search provide the file
    name with '*' as wildcard(s). Returns None is not found.

    :param dir_path: string for the input directory path
    :param file_search: string with a * wildcard for the file being searched for.
    :return: string with the path to the file

    .. code:: python

        import rsgislib.tools.filetools
        file_path = rsgislib.tools.filetools.find_file_none("in/dir", "*N15W093*.tif")
        if file_path is not None:
            print(file_path)

    """
    files = glob.glob(os.path.join(dir_path, file_search))
    if len(files) != 1:
        return None
    return files[0]


def find_files_ext(dir_path: str, ending: str) -> dict:
    """
    Find all the files within a directory structure with a specific file ending.
    The files are return as dictionary using the file name as the dictionary key.
    This means you cannot have files with the same name within the structure.

    :param dir_path: the base directory path within which to search.
    :param ending: the file ending (e.g., .txt, or txt or .kea, kea).
    :return: dict with file name as key

    .. code:: python

        import rsgislib.tools.filetools
        file_paths = rsgislib.tools.filetools.find_files_ext("in/dir", ".tif")

    """
    out_file_dict = dict()
    for root, dirs, files in os.walk(dir_path):
        for file in files:
            if file.endswith(ending):
                file_found = os.path.join(root, file)
                if os.path.isfile(file_found):
                    out_file_dict[file] = file_found
    return out_file_dict


def find_files_mpaths_ext(dir_paths: list, ending: str) -> dict:
    """
    Find all the files within a list of input directories and the structure beneath
    with a specific file ending. The files are return as dictionary using the file
    name as the dictionary key. This means you cannot have files with the same name
    within the structure.

    :param dir_paths: a list of base directory paths within which to search.
    :param ending: the file ending (e.g., .txt, or txt or .kea, kea).
    :return: dict with file name as key

    .. code:: python

        import rsgislib.tools.filetools
        dir_paths = ["in/dir", "test/dir", "img/files"]
        file_paths = rsgislib.tools.filetools.find_files_mpaths_ext(dir_paths, ".tif")


    """
    out_file_dict = dict()
    for dir_path in dir_paths:
        for root, dirs, files in os.walk(dir_path):
            for file in files:
                if file.endswith(ending):
                    file_found = os.path.join(root, file)
                    if os.path.isfile(file_found):
                        out_file_dict[file] = file_found
    return out_file_dict


def find_first_file(dir_path: str, file_search: str, rtn_except: bool = True) -> str:
    """
    Search for a single file with a path using glob. Therefore, the file
    path returned is a true path. Within the file_search provide the file
    name with '*' as wildcard(s).

    :param dir_path: The directory within which to search, note that the search will
                     be within sub-directories within the base directory until a file
                     meeting the search criteria are met.
    :param file_search: The file search string in the file name and must contain a
                        wild character (i.e., *).
    :param rtn_except: if True then an exception will be raised if no file or multiple
                       files are found (default). If False then None will be returned
                       rather than an exception raised.
    :return: The file found (or None if rtn_except=False)

    .. code:: python

        import rsgislib.tools.filetools
        file_paths = rsgislib.tools.filetools.find_first_file("in/dir", "*N15W093*.tif")

    """
    files = None
    for root, dirs, files in os.walk(dir_path):
        files = glob.glob(os.path.join(root, file_search))
        if len(files) > 0:
            break
    out_file = None
    if (files is not None) and (len(files) == 1):
        out_file = files[0]
    elif rtn_except:
        raise rsgislib.RSGISPyException(
            "Could not find a single file ({0}) in {1}; "
            "found {2} files.".format(file_search, dir_path, len(files))
        )
    return out_file


def get_files_mod_time(
    file_lst: list,
    dt_before: datetime.datetime = None,
    dt_after: datetime.datetime = None,
) -> list:
    """
    A function which subsets a list of files based on datetime of
    last modification. The function also does a check as to whether
    a file exists, files which don't exist will be ignored.

    :param file_lst: The list of file path - represented as strings.
    :param dt_before: a datetime object with a date/time where files modified
                      before this will be returned
    :param dt_after: a datetime object with a date/time where files modified
                     after this will be returned

    Example:

    .. code:: python

        import glob
        import datetime
        import rsgislib.tools.filetools

        input_files = glob.glob("in/dir/*.tif")
        dt_before = datetime.datetime(year=2020, month=12, day=25, hour=12, minute=30)
        file_path = rsgislib.tools.filetools.get_files_mod_time(input_files, dt_before)

    """
    if (dt_before is None) and (dt_after is None):
        raise rsgislib.RSGISPyException(
            "You must define at least one of dt_before or dt_after"
        )
    out_file_lst = list()
    for cfile in file_lst:
        if os.path.exists(cfile):
            mod_time_stamp = os.path.getmtime(cfile)
            mod_time = datetime.datetime.fromtimestamp(mod_time_stamp)
            if (dt_before is not None) and (mod_time < dt_before):
                out_file_lst.append(cfile)
            if (dt_after is not None) and (mod_time > dt_after):
                out_file_lst.append(cfile)
    return out_file_lst


def find_files_size_limits(
    dir_path: str, file_search: str, min_size: int = 0, max_size: int = None
) -> list:
    """
    Search for files with a path using glob. Therefore, the file
    paths returned is a true path. Within the file_search provide the file
    names with '*' as wildcard(s).

    :param dir_path: string for the input directory path
    :param file_search: string with a * wildcard for the file being searched for.
    :param min_size: the minimum file size in bytes (default is 0)
    :param max_size: the maximum file size in bytes, if None (default) then ignored.
    :return: string with the path to the file

    Example:

    .. code:: python

        import rsgislib.tools.filetools
        file_paths = rsgislib.tools.filetools.find_files_size_limits("in/dir",
                                                                     "*N15W093*.tif",
                                                                     0, 100000)

    """
    files = glob.glob(os.path.join(dir_path, file_search))
    out_files = list()
    for c_file in files:
        file_size = get_file_size(c_file)
        if (max_size is None) and (file_size > min_size):
            out_files.append(c_file)
        elif (file_size > min_size) and (file_size < max_size):
            out_files.append(c_file)
    return out_files


def is_cmd_tool_avail(cmd_name: str, test_call_cmd: List[str] = None):
    """
    A function which finds if an executable command is available
    on the system path. If you use the option to call the command
    then care is needed as this is running a command on the system
    and this function does no checking on what command and options
    has been past(!).

    :param cmd_name: the name of the command to test (e.g., tar, gzip etc.)
    :param test_call_cmd: Option to call the command to test if an error
                          is thrown. Default is None and then command is not
                          called but if a List (e.g., ["tar", "-h"]) is provided
                          then it will be executed to test whether the command
                          runs without throwing an error. Note, this is
    :return: boolean (True: Command available. False Command not available)
    """
    command_exists = shutil.which(cmd_name) is not None
    if (test_call_cmd is not None) and command_exists:
        import subprocess

        if not isinstance(test_call_cmd, list):
            raise rsgislib.RSGISPyException(
                "Input test command was not provided as a list."
            )

        try:
            rtn_info = subprocess.run(
                test_call_cmd,
                capture_output=True,
                text=True,
                check=True,
            )
            if rtn_info.returncode == 0:
                return True
            else:
                return False
        except subprocess.CalledProcessError:
            return False
        except Exception:
            return False
    return command_exists


def file_is_hidden(dir_path: str) -> bool:
    """
    A function to test whether a file or folder is 'hidden' or not on the
    file system. Should be cross platform between Linux/UNIX and windows.

    :param dir_path: input file path to be tested
    :return: boolean (True = hidden)

    Example:

    .. code:: python

        import rsgislib.tools.filetools
        if rsgislib.tools.filetools.file_is_hidden("in/dir/img.kea"):
            print("File is hidden")

    """
    dir_path = os.path.abspath(dir_path)
    if os.name == "nt":
        import win32api
        import win32con

        attribute = win32api.GetFileAttributes(dir_path)
        return attribute & (
            win32con.FILE_ATTRIBUTE_HIDDEN | win32con.FILE_ATTRIBUTE_SYSTEM
        )
    else:
        file_name = os.path.basename(dir_path)
        return file_name.startswith(".")


def get_dir_list(dir_path: str, inc_hidden: bool = False) -> list:
    """
    Function which get the list of directories within the specified path.

    :param dir_path: file path to search within
    :param inc_hidden: boolean specifying whether hidden files should be
                       included (default=False)
    :return: list of directory paths

    Example:

    .. code:: python

        import rsgislib.tools.filetools
        files = rsgislib.tools.filetools.get_dir_list("in/dir")

    """
    out_dir_lst = list()
    dir_listing = os.listdir(dir_path)
    for item in dir_listing:
        c_path = os.path.join(dir_path, item)
        if os.path.isdir(c_path):
            if not inc_hidden:
                if not file_is_hidden(c_path):
                    out_dir_lst.append(c_path)
            else:
                out_dir_lst.append(c_path)
    return out_dir_lst


def convert_file_size_units(in_size: int, in_unit: str, out_unit: str) -> float:
    """
    A function which converts between file size units

    :param in_size: input file size
    :param in_unit: the input unit for the file size. Options: bytes, kb, mb, gb, tb
    :param out_unit: the output unit for the file size. Options: bytes, kb, mb, gb, tb
    :return: float for the output file size

    """
    in_unit = in_unit.lower()
    if in_unit not in ["bytes", "kb", "mb", "gb", "tb"]:
        raise rsgislib.RSGISPyException(
            "Input unit must be one of: bytes, kb, mb, gb, tb"
        )

    out_unit = out_unit.lower()
    if out_unit not in ["bytes", "kb", "mb", "gb", "tb"]:
        raise rsgislib.RSGISPyException(
            "Output unit must be one of: bytes, kb, mb, gb, tb"
        )

    if in_unit == "bytes":
        file_size_bytes = in_size
    elif in_unit == "kb":
        file_size_bytes = in_size * 1024.0
    elif in_unit == "mb":
        file_size_bytes = in_size * (1024.0**2)
    elif in_unit == "gb":
        file_size_bytes = in_size * (1024.0**3)
    elif in_unit == "tb":
        file_size_bytes = in_size * (1024.0**4)
    else:
        raise rsgislib.RSGISPyException("Input unit it not recognised.")

    if out_unit == "bytes":
        out_file_size = file_size_bytes
    elif out_unit == "kb":
        out_file_size = file_size_bytes / 1024.0
    elif out_unit == "mb":
        out_file_size = file_size_bytes / (1024.0**2)
    elif out_unit == "gb":
        out_file_size = file_size_bytes / (1024.0**3)
    elif out_unit == "tb":
        out_file_size = file_size_bytes / (1024.0**4)
    else:
        raise rsgislib.RSGISPyException("Output unit it not recognised.")

    return out_file_size


def get_file_size(file_path: str, unit: str = "bytes") -> float:
    """
    A function which returns the file size of a file in the specified unit.

    Units:
    * bytes
    * kb - kilobytes  (bytes / 1024)
    * mb - megabytes  (bytes / 1024^2)
    * gb - gigabytes  (bytes / 1024^3)
    * tb - terabytes  (bytes / 1024^4)

    :param file_path: the path to the file for which the size is to be calculated.
    :param unit: the unit for the file size. Options: bytes, kb, mb, gb, tb
    :return: float for the file size.

    """
    import pathlib

    unit = unit.lower()
    if unit not in ["bytes", "kb", "mb", "gb", "tb"]:
        raise rsgislib.RSGISPyException("Unit must be one of: bytes, kb, mb, gb, tb")

    p = pathlib.Path(file_path)
    if p.exists() and p.is_file():
        file_size_bytes = p.stat().st_size
        out_file_size = convert_file_size_units(file_size_bytes, "bytes", unit)
    else:
        raise rsgislib.RSGISPyException("Input file path does not exist")
    return out_file_size


def get_file_lock(
    input_file: str,
    sleep_period: int = 1,
    wait_iters: int = 120,
    use_except: bool = False,
) -> bool:
    """
    A function which gets a lock on a file.

    The lock file will be a unix hidden file (i.e., starts with a .) and it will
    have .lok added to the end. E.g., for input file hello_world.txt the lock file
    will be .hello_world.txt.lok. The contents of the lock file will be the time and
    date of creation.

    Using the default parameters (sleep 1 second and wait 120 iterations) if the
    lock isn't available it will be retried every second for 120 seconds (i.e., 2 mins).

    :param input_file: The input file for which the lock will be created.
    :param sleep_period: time in seconds to sleep for, if the lock isn't
                         available. (Default=1 second)
    :param wait_iters: the number of iterations to wait for before giving
                       up. (Default=120)
    :param use_except: Boolean. If True then an exception will be thrown if the lock
                       is not available. If False (default) False will be returned
                       if the lock is not successful.
    :return: boolean. True: lock was successfully gained. False: lock was not gained.

    """
    file_path, file_name = os.path.split(input_file)
    lock_file_name = ".{}.lok".format(file_name)
    lock_file_path = os.path.join(file_path, lock_file_name)

    got_lock = False
    for i in range(wait_iters + 1):
        if not os.path.exists(lock_file_path):
            got_lock = True
            break
        time.sleep(sleep_period)

    if got_lock:
        c_datetime = datetime.datetime.now()
        f = open(lock_file_path, "w")
        f.write("{}\n".format(c_datetime.isoformat()))
        f.flush()
        f.close()
    elif use_except:
        raise rsgislib.RSGISPyException(
            "Lock could not be gained for file: {}".format(input_file)
        )

    return got_lock


def release_file_lock(input_file: str):
    """
    A function which releases a lock file for the input file.

    :param input_file: The input file for which the lock will be created.

    """
    file_path, file_name = os.path.split(input_file)
    lock_file_name = ".{}.lok".format(file_name)
    lock_file_path = os.path.join(file_path, lock_file_name)
    if os.path.exists(lock_file_path):
        if not delete_file_silent(lock_file_path):
            raise rsgislib.RSGISPyException(
                "Could not delete the lock file... something has gone wrong!"
            )


def clean_file_locks(dir_path: str, timeout: int = 3600):
    """
    A function which cleans up any remaining lock file (i.e., if an application
    has crashed). The timeout time will be compared with the time written within
    the file.

    :param dir_path: the file path to search for lock files (i.e., ".*.lok")
    :param timeout: the time (in seconds) for the timeout. Default: 3600 (1 hours)

    """
    import rsgislib.tools.utils

    c_dateime = datetime.datetime.now()
    lock_files = glob.glob(os.path.join(dir_path, ".*.lok"))
    for lock_file_path in lock_files:
        create_date_str = rsgislib.tools.utils.read_text_file_no_new_lines(
            lock_file_path
        )
        create_date = datetime.datetime.fromisoformat(create_date_str)
        time_since_create = (c_dateime - create_date).total_seconds()
        if time_since_create > timeout:
            if not delete_file_silent(lock_file_path):
                raise rsgislib.RSGISPyException(
                    "Could not delete the lock file... something has gone wrong!"
                )


def sort_imgs_to_dirs_utm(input_imgs_dir: str, file_search_str: str, out_base_dir: str):
    """
    A function which will sort a series of input image files which
    a projected using the UTM system into individual directories per
    UTM zone. Please note that the input files are moved on your system!!

    :param input_imgs_dir: directory where the input files are to be found.
    :param file_search_str: the wildcard search string to find files within
                            the input directory (e.g., "in_dir/*.kea").
    :param out_base_dir: the output directory where the UTM folders will be created
                         and the files copied.

    """
    import rsgislib.imageutils

    in_files = glob.glob(os.path.join(input_imgs_dir, file_search_str))
    for img_file in in_files:
        utm_zone = rsgislib.imageutils.getUTMZone(img_file)
        if utm_zone is not None:
            out_dir = os.path.join(out_base_dir, "utm" + utm_zone)
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            img_file_list = rsgislib.imageutils.getImageFiles(img_file)
            for tmp_file in img_file_list:
                print("Moving: " + tmp_file)
                outFile = os.path.join(out_dir, os.path.basename(tmp_file))
                shutil.move(tmp_file, outFile)


def create_directory_archive(in_dir: str, out_arch: str, arch_format: str) -> str:
    """
    A function which creates an archive from an input directory. This function uses
    subprocess to call the appropriate command line function.

    Please note that this function has similar functionality to shutil.make_archive
    and I would recommend you use that but I found it sometimes produces an error so
    I provided this function which uses the terminal functions as a drop in replacement.

    :param in_dir: The input directory path for which the archive with be created.
    :param out_arch: The output archive file path and name. Note this should not
                     include an extension as this will be added automatically.
    :param arch_format: The format for the archive. The options are: zip, tar,
                        gztar, bztar, xztar
    :return: a string with the full file path and name, including the file extension.

    """
    import subprocess

    if not is_path_valid(in_dir):
        raise rsgislib.RSGISPyException(f"The input directory is not valid: {in_dir}")

    c_pwd = os.getcwd()
    out_arch = os.path.abspath(out_arch)
    in_dir = os.path.abspath(in_dir)
    base_dir = os.path.dirname(in_dir)
    dir_name = os.path.split(in_dir)[1]
    os.chdir(base_dir)

    if arch_format == "zip":
        out_arch_file = "{}.zip".format(out_arch.strip())
        if not does_path_exists_or_creatable(out_arch_file):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {out_arch_file}"
            )
        cmd = ["zip", "-r", out_arch_file, dir_name]
        subprocess.run(cmd, check=True)
    elif arch_format == "tar":
        out_arch_file = "{}.tar".format(out_arch.strip())
        if not does_path_exists_or_creatable(out_arch_file):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {out_arch_file}"
            )
        cmd = ["tar", "-cvf", out_arch_file, dir_name]
        subprocess.run(cmd, check=True)
    elif arch_format == "gztar":
        out_arch_file = "{}.tar.gz".format(out_arch.strip())
        if not does_path_exists_or_creatable(out_arch_file):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {out_arch_file}"
            )
        cmd = ["tar", "-cvzf", out_arch_file, dir_name]
        subprocess.run(cmd, check=True)
    elif arch_format == "bztar":
        out_arch_file = "{}.tar.bz2".format(out_arch.strip())
        if not does_path_exists_or_creatable(out_arch_file):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {out_arch_file}"
            )
        cmd = ["tar", "-cvjSf", out_arch_file, dir_name]
        subprocess.run(cmd, check=True)
    elif arch_format == "xztar":
        out_arch_file = "{}.tar.xz".format(out_arch.strip())
        if not does_path_exists_or_creatable(out_arch_file):
            raise rsgislib.RSGISPyException(
                f"Output file path is not creatable: {out_arch_file}"
            )
        cmd = ["tar", "-cvJf", out_arch_file, dir_name]
        subprocess.run(cmd, check=True)
    else:
        raise rsgislib.RSGISPyException("Do not recognise the archive format specifed.")

    os.chdir(c_pwd)
    return out_arch_file


def create_sha1_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA1 hash string of the input file.

    :param input_file: the input file for which the SHA1 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA1 hash string of the file.

    """
    import hashlib

    sha1_hash = hashlib.sha1()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha1_hash.update(byte_block)
    return sha1_hash.hexdigest()


def create_sha224_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA224 hash string of the input file.

    :param input_file: the input file for which the SHA224 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA224 hash string of the file.

    """
    import hashlib

    sha224_hash = hashlib.sha224()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha224_hash.update(byte_block)
    return sha224_hash.hexdigest()


def create_sha256_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA256 hash string of the input file.

    :param input_file: the input file for which the SHA256 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA256 hash string of the file.

    """
    import hashlib

    sha256_hash = hashlib.sha256()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def create_sha384_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA384 hash string of the input file.

    :param input_file: the input file for which the SHA384 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA384 hash string of the file.

    """
    import hashlib

    sha384_hash = hashlib.sha384()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha384_hash.update(byte_block)
    return sha384_hash.hexdigest()


def create_sha512_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA512 hash string of the input file.

    :param input_file: the input file for which the SHA512 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA512 hash string of the file.

    """
    import hashlib

    sha512_hash = hashlib.sha512()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha512_hash.update(byte_block)
    return sha512_hash.hexdigest()


def create_md5_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the MD5 hash string of the input file.

    :param input_file: the input file for which the MD5 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: MD5 hash string of the file.

    """
    import hashlib

    md5_hash = hashlib.md5()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            md5_hash.update(byte_block)
    return md5_hash.hexdigest()


def create_blake2b_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the Blake2B hash string of the input file.

    :param input_file: the input file for which the Blake2B hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: Blake2B hash string of the file.

    """
    import hashlib

    blake2b_hash = hashlib.blake2b()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            blake2b_hash.update(byte_block)
    return blake2b_hash.hexdigest()


def create_blake2s_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the Blake2S hash string of the input file.

    :param input_file: the input file for which the Blake2S hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: Blake2S hash string of the file.

    """
    import hashlib

    blake2s_hash = hashlib.blake2s()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            blake2s_hash.update(byte_block)
    return blake2s_hash.hexdigest()


def create_sha3_224_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA3_224 hash string of the input file.

    :param input_file: the input file for which the SHA3_224 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA3_224 hash string of the file.

    """
    import hashlib

    sha224_hash = hashlib.sha3_224()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha224_hash.update(byte_block)
    return sha224_hash.hexdigest()


def create_sha3_256_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA3_256 hash string of the input file.

    :param input_file: the input file for which the SHA3_256 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA3_256 hash string of the file.

    """
    import hashlib

    sha256_hash = hashlib.sha3_256()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def create_sha3_384_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA3_384 hash string of the input file.

    :param input_file: the input file for which the SHA3_384 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA3_384 hash string of the file.

    """
    import hashlib

    sha384_hash = hashlib.sha3_384()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha384_hash.update(byte_block)
    return sha384_hash.hexdigest()


def create_sha3_512_hash(input_file: str, block_size: int = 4096) -> str:
    """
    A function which calculates finds the SHA3_512 hash string of the input file.

    :param input_file: the input file for which the SHA3_512 hash string with be found.
    :param block_size: the size of the blocks the file is read in in bytes
                       (default 4096; i.e., 4kb)
    :return: SHA3_512 hash string of the file.

    """
    import hashlib

    sha512_hash = hashlib.sha3_512()
    with open(input_file, "rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(block_size), b""):
            sha512_hash.update(byte_block)
    return sha512_hash.hexdigest()


def untar_file(
    in_file: str, out_dir: str, gen_arch_dir: bool = True, verbose: bool = False
) -> str:
    """
    A function which extracts data from a tar file into the specified
    output directory. Optionally, an output directory of the same name
    as the archive file can be created for the output files.

    :param in_file: The input archive file.
    :param out_dir: The output directory which must exist (if gen_arch_dir=True then
                    a new directory will be created within the out_dir
    :param gen_arch_dir: Create a new directory with the same name as the input file
                         where the output files will be extracted to. (Default: True)
    :param verbose: If True (default: False) then more user feedback will be printed
                    to the console.
    :return: output directory where data was extracted to.

    """
    if not os.path.exists(out_dir):
        raise rsgislib.RSGISPyException("Output directory does not exist.")
    if not os.path.exists(in_file):
        raise rsgislib.RSGISPyException("Input file does not exist.")
    in_file = os.path.abspath(in_file)
    out_dir = os.path.abspath(out_dir)
    process_dir = out_dir
    if gen_arch_dir:
        basename = get_file_basename(in_file)
        process_dir = os.path.join(out_dir, basename)
    if not os.path.exists(process_dir):
        os.makedirs(process_dir)
    c_dir = os.getcwd()
    os.chdir(process_dir)
    if verbose:
        print("Extracting: {}".format(in_file))
        print("Output to: {}".format(process_dir))
        cmd = ["tar", "-xvf", in_file]
    else:
        cmd = ["tar", "-xf", in_file]
    try:
        import subprocess

        subprocess.run(cmd, check=True)
    except OSError as e:
        os.chdir(c_dir)
        raise rsgislib.RSGISPyException("Could not extract data: {}".format(cmd))
    os.chdir(c_dir)
    return process_dir


def untar_gz_file(
    in_file: str, out_dir: str, gen_arch_dir: bool = True, verbose: bool = False
) -> str:
    """
    A function which extracts data from a tar.gz file into the specified
    output directory. Optionally, an output directory of the same name
    as the archive file can be created for the output files.

    :param in_file: The input archive file.
    :param out_dir: The output directory which must exist (if gen_arch_dir=True then
                    a new directory will be created within the out_dir
    :param gen_arch_dir: Create a new directory with the same name as the input file
                         where the output files will be extracted to. (Default: True)
    :param verbose: If True (default: False) then more user feedback will be printed
                    to the console.
    :return: output directory where data was extracted to.

    """
    if not os.path.exists(out_dir):
        raise rsgislib.RSGISPyException("Output directory does not exist.")
    if not os.path.exists(in_file):
        raise rsgislib.RSGISPyException("Input file does not exist.")
    in_file = os.path.abspath(in_file)
    out_dir = os.path.abspath(out_dir)
    process_dir = out_dir
    if gen_arch_dir:
        basename = get_file_basename(in_file)
        process_dir = os.path.join(out_dir, basename)
    if not os.path.exists(process_dir):
        os.makedirs(process_dir)
    c_dir = os.getcwd()
    os.chdir(process_dir)
    if verbose:
        print("Extracting: {}".format(in_file))
        print("Output to: {}".format(process_dir))
        cmd = ["tar", "-xvzf", in_file]
        print(cmd)
    else:
        cmd = ["tar", "-xzf", in_file]
    try:
        import subprocess

        subprocess.run(cmd, check=True)
    except OSError as e:
        os.chdir(c_dir)
        raise rsgislib.RSGISPyException("Could not extract data: {}".format(cmd))
    os.chdir(c_dir)
    return process_dir


def unzip_file(
    in_file: str, out_dir: str, gen_arch_dir: bool = True, verbose: bool = False
) -> str:
    """
    A function which extracts data from a zip file into the specified
    output directory. Optionally, an output directory of the same name
    as the archive file can be created for the output files.

    :param in_file: The input archive file.
    :param out_dir: The output directory which must exist (if gen_arch_dir=True then
                    a new directory will be created within the out_dir
    :param gen_arch_dir: Create a new directory with the same name as the input file
                         where the output files will be extracted to. (Default: True)
    :param verbose: If True (default: False) then more user feedback will be printed
                    to the console.
    :return: output directory where data was extracted to.

    """
    if not os.path.exists(out_dir):
        raise rsgislib.RSGISPyException("Output directory does not exist.")
    if not os.path.exists(in_file):
        raise rsgislib.RSGISPyException("Input file does not exist.")
    in_file = os.path.abspath(in_file)
    out_dir = os.path.abspath(out_dir)
    process_dir = out_dir
    if gen_arch_dir:
        basename = get_file_basename(in_file)
        process_dir = os.path.join(out_dir, basename)
    if not os.path.exists(process_dir):
        os.makedirs(process_dir)
    c_dir = os.getcwd()
    os.chdir(process_dir)
    cmd = ["unzip", in_file]
    if verbose:
        print("Extracting: {}".format(in_file))
        print("Output to: {}".format(process_dir))
        print(cmd)
    try:
        import subprocess

        subprocess.run(cmd, check=True)
    except OSError as e:
        os.chdir(c_dir)
        raise rsgislib.RSGISPyException("Could not extract data: {}".format(cmd))
    os.chdir(c_dir)
    return process_dir


def untar_bz_file(
    in_file: str, out_dir: str, gen_arch_dir: bool = True, verbose: bool = False
) -> str:
    """
    A function which extracts data from a tar.bz file into the specified
    output directory. Optionally, an output directory of the same name
    as the archive file can be created for the output files.

    :param in_file: The input archive file.
    :param out_dir: The output directory which must exist (if gen_arch_dir=True then
                    a new directory will be created within the out_dir
    :param gen_arch_dir: Create a new directory with the same name as the input file
                         where the output files will be extracted to. (Default: True)
    :param verbose: If True (default: False) then more user feedback will be printed
                    to the console.
    :return: output directory where data was extracted to.

    """
    if not os.path.exists(out_dir):
        raise rsgislib.RSGISPyException("Output directory does not exist.")
    if not os.path.exists(in_file):
        raise rsgislib.RSGISPyException("Input file does not exist.")
    in_file = os.path.abspath(in_file)
    out_dir = os.path.abspath(out_dir)
    process_dir = out_dir
    if gen_arch_dir:
        basename = get_file_basename(in_file)
        process_dir = os.path.join(out_dir, basename)
    if not os.path.exists(process_dir):
        os.makedirs(process_dir)
    c_dir = os.getcwd()
    os.chdir(process_dir)
    if verbose:
        print("Extracting: {}".format(in_file))
        print("Output to: {}".format(process_dir))
        cmd = ["tar", "-xvjf", in_file]
        print(cmd)
    else:
        cmd = ["tar", "-xjf", in_file]
    try:
        import subprocess

        subprocess.run(cmd, check=True)
    except OSError as e:
        os.chdir(c_dir)
        raise rsgislib.RSGISPyException("Could not extract data: {}".format(cmd))
    os.chdir(c_dir)
    return process_dir


def create_targz_arch(out_arch_file: str, file_list: list, base_path: str = None):
    """
    A function which can be used to create a tar.gz file containing the
    list of input files. If you wish to remove some of the directory
    structure from the file paths in provided then a single base_path
    can be provided and will be removed from the file paths in the archive.

    :param out_arch_file: the output tar.gz file path
    :param file_list: the list of files to be added to the archive.
    :param base_path: the base path which will be removed from all the input files.
                      Note, this means all the input files must have the same base
                      path. Optional: Default is None (i.e., ignored).

    """
    import tarfile

    out_arch_file = os.path.abspath(out_arch_file)

    if base_path is not None:
        cwd = os.getcwd()
        os.chdir(base_path)

    with tarfile.open(out_arch_file, "w:gz") as tar_out_file:
        for c_file in file_list:
            if base_path is not None:
                c_file = os.path.relpath(c_file, base_path)
            tar_out_file.add(c_file)

    if base_path is not None:
        os.chdir(cwd)



import os
import pytest

SKLEARN_NOT_AVAIL = False
try:
    import sklearn
except ImportError:
    SKLEARN_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
TIMESERIES_DATA_DIR = os.path.join(DATA_DIR, "timeseries")


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_get_ST_model_coeffs(tmp_path):
    from rsgislib.timeseries import modelfitting
    import rsgislib.tools.utils

    sen2_imgs_path = os.path.join(TIMESERIES_DATA_DIR, "sen2_subs")
    sen2_imgs_lut_ref_file = os.path.join(TIMESERIES_DATA_DIR, "./timeseries_imgs.json")
    sen2_imgs_lut_file = os.path.join(tmp_path, "./timeseries_imgs.json")

    sen2_imgs_ref_dict = rsgislib.tools.utils.read_json_to_dict(sen2_imgs_lut_ref_file)
    sen2_imgs_dict = dict()
    for date_key in sen2_imgs_ref_dict:
        sen2_imgs_dict[date_key] = os.path.join(
            sen2_imgs_path, sen2_imgs_ref_dict[date_key]
        )

    rsgislib.tools.utils.write_dict_to_json(sen2_imgs_dict, sen2_imgs_lut_file)

    coeffs_img = os.path.join(tmp_path, "timeseries_mdl_coeffs.kea")
    modelfitting.get_ST_model_coeffs(
        sen2_imgs_lut_file, coeffs_img, bands=[3, 7, 9], num_processes=1
    )

    assert os.path.exists(coeffs_img)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_predict_for_date(tmp_path):
    from rsgislib.timeseries import modelfitting

    coeffs_img = os.path.join(TIMESERIES_DATA_DIR, "timeseries_mdl_coeffs.kea")
    output_img = os.path.join(tmp_path, "sen2_pred_img_20190610.kea")
    modelfitting.predict_for_date("2019-06-10", coeffs_img, output_img)

    assert os.path.exists(output_img)



#! /usr/bin/env python
"""
The tools.utm module contains some useful tools for work with UTM projection.

This code has been adapted from https://github.com/Turbo87/utm/blob/master/utm/conversion.py

"""
import numpy

import rsgislib

K0 = 0.9996

E = 0.00669438
E2 = E * E
E3 = E2 * E
E_P2 = E / (1.0 - E)

SQRT_E = numpy.sqrt(1 - E)
_E = (1 - SQRT_E) / (1 + SQRT_E)
_E2 = _E * _E
_E3 = _E2 * _E
_E4 = _E3 * _E
_E5 = _E4 * _E

M1 = 1 - E / 4 - 3 * E2 / 64 - 5 * E3 / 256
M2 = 3 * E / 8 + 3 * E2 / 32 + 45 * E3 / 1024
M3 = 15 * E2 / 256 + 45 * E3 / 1024
M4 = 35 * E3 / 3072

P2 = 3.0 / 2 * _E - 27.0 / 32 * _E3 + 269.0 / 512 * _E5
P3 = 21.0 / 16 * _E2 - 55.0 / 32 * _E4
P4 = 151.0 / 96 * _E3 - 417.0 / 128 * _E5
P5 = 1097.0 / 512 * _E4

R = 6378137

ZONE_LETTERS = "CDEFGHJKLMNPQRSTUVWXX"


def min_max_eastings_for_lat(latitude: float, zone: int = 1) -> (float, float):
    """
    Find the minimum and maximum Eastings for a specified latitude within a specific zone.
    This should be the same for all zones.

    :param latitude: float
    :param zone: int
    :return: tuple [min, max]
    """
    if isinstance(latitude, numpy.ndarray):
        raise rsgislib.RSGISPyException("Don't input an array.")

    central_long = zone_number_to_central_longitude(zone)
    min_long = central_long - 3.0
    max_long = central_long + 3.0

    min_utm = from_latlon(latitude, min_long, force_zone_number=zone)
    max_utm = from_latlon(latitude, max_long, force_zone_number=zone)

    return min_utm[0], max_utm[0]


def epsg_for_utm(zone: int, hemisphere: str) -> int:
    """
    Return EPSG code for given UTM zone and hemisphere using WGS84 datum.

    :param zone: UTM zone
    :param hemisphere: hemisphere either 'N' or 'S'
    :return: corresponding EPSG code
    """
    if hemisphere not in ["N", "S"]:
        raise rsgislib.RSGISPyException('Invalid hemisphere ("N" or "S").')

    if zone < 0 or zone > 60:
        raise rsgislib.RSGISPyException("UTM zone ouside valid range.")

    if hemisphere == "N":
        ns = 600
    else:
        ns = 700

    if zone == 0:
        zone = 61

    return int(32000 + ns + zone)


def utm_from_epsg(epsg_code: int) -> (int, str):
    """
    Return UTM zone and hemisphere from a EPSG code using WGS84 datum.

    :param epsg_code: epsg code for the UTM projection.
    :return: zone, hemisphere
    """
    h_zone = epsg_code - 32000

    if h_zone < 700:
        hemisphere = "N"
        zone = h_zone - 600
    else:
        hemisphere = "S"
        zone = h_zone - 700

    return zone, hemisphere


def split_wgs84_bbox_utm_zones(wgs84_bbox):
    """
    Function which splits a bounding box (MinX, MaxX, MinY, MaxY) into
    a number of small bounding boxes (MinX, MaxX, MinY, MaxY) at UTM zone
    boundaries.

    :param wgs84_bbox: bounding box in lat/long WGS84 (MinX, MaxX, MinY, MaxY)
    :return: list [utm zone, bbox]
    """
    utm_tl = from_latlon(wgs84_bbox[3], wgs84_bbox[0])
    utm_tr = from_latlon(wgs84_bbox[3], wgs84_bbox[1])
    utm_br = from_latlon(wgs84_bbox[2], wgs84_bbox[1])
    utm_bl = from_latlon(wgs84_bbox[2], wgs84_bbox[0])

    utm_zones = []
    for utm_coord in [utm_tl, utm_tr, utm_br, utm_bl]:
        utm_zones.append(utm_coord[2])

    min_zone = numpy.min(utm_zones)
    max_zone = numpy.max(utm_zones)

    zones = numpy.arange(min_zone, max_zone + 1, 1)

    out_bboxs = []
    for zone in zones:
        zone_min = ((zone * 6.0) - 6.0) - 180.0
        zone_max = (zone * 6.0) - 180.0

        # Intersect input BBOX with zone bound
        x_min = zone_min
        if wgs84_bbox[0] > zone_min:
            x_min = wgs84_bbox[0]

        x_max = zone_max
        if wgs84_bbox[1] < zone_max:
            x_max = wgs84_bbox[1]

        out_bboxs.append([zone, [x_min, x_max, wgs84_bbox[2], wgs84_bbox[3]]])

    return out_bboxs


def check_valid_zone(
    zone_number: int, zone_letter: str = None, use_exp: bool = False
) -> bool:
    """
    Check that the UTM zone is valid.

    :param zone_number: int for the UTM zone
    :param zone_letter: string for the zone letter.
    :param use_exp: If True then an exception will be thrown if the zone is invalid
                    if False (default) then a boolean value of False will be returned
                    by the function if zone is not valid.
    :return: True if zone is valid and Falsei if not.

    """
    valid_zone = True
    if not 1 <= zone_number <= 60:
        if use_exp:
            raise rsgislib.RSGISPyException(
                "zone number out of range (must be between 1 and 60)"
            )
        valid_zone = False

    if zone_letter:
        zone_letter = zone_letter.upper()

        if not "C" <= zone_letter <= "X" or zone_letter in ["I", "O"]:
            if use_exp:
                raise rsgislib.RSGISPyException(
                    "zone letter out of range (must be between C and X)"
                )
            valid_zone = False

    return valid_zone


def to_latlon(
    easting, northing, zone_number, zone_letter=None, northern=None, strict=True
):
    """
    This function convert an UTM coordinate into Latitude and Longitude

    :param easting: int or array: Easting value of UTM coordinate
    :param northing: int or array: Northing value of UTM coordinate
    :param zone number: int: Zone Number is represented with global map numbers of an UTM Zone
        Numbers Map. More information see http://www.jaworski.ca/utmzones.htm
    :param zone_letter: str: Zone Letter can be represented as string values.
        Where UTM Zone Designators can be accessed in http://www.jaworski.ca/utmzones.htm
    :param northern: bool: You can set True or False to set this parameter. Default is None
    """
    import rsgislib.tools.utils

    if not zone_letter and northern is None:
        raise ValueError("either zone_letter or northern needs to be set")

    elif zone_letter and northern is not None:
        raise ValueError("set either zone_letter or northern, but not both")

    if strict:
        if not rsgislib.tools.utils.in_bounds(
            easting, 100000, 1000000, upper_strict=True
        ):
            raise rsgislib.RSGISPyException(
                "easting out of range (must be between 100.000 m and 999.999 m)"
            )
        if not rsgislib.tools.utils.in_bounds(northing, 0, 10000000):
            raise rsgislib.RSGISPyException(
                "northing out of range (must be between 0 m and 10.000.000 m)"
            )

    check_valid_zone(zone_number, zone_letter)

    if zone_letter:
        zone_letter = zone_letter.upper()
        northern = zone_letter >= "N"

    x = easting - 500000
    y = northing

    if not northern:
        y -= 10000000

    m = y / K0
    mu = m / (R * M1)

    p_rad = (
        mu
        + P2 * numpy.sin(2 * mu)
        + P3 * numpy.sin(4 * mu)
        + P4 * numpy.sin(6 * mu)
        + P5 * numpy.sin(8 * mu)
    )

    p_sin = numpy.sin(p_rad)
    p_sin2 = p_sin * p_sin

    p_cos = numpy.cos(p_rad)

    p_tan = p_sin / p_cos
    p_tan2 = p_tan * p_tan
    p_tan4 = p_tan2 * p_tan2

    ep_sin = 1 - E * p_sin2
    ep_sin_sqrt = numpy.sqrt(1 - E * p_sin2)

    n = R / ep_sin_sqrt
    r = (1 - E) / ep_sin

    c = _E * p_cos**2
    c2 = c * c

    d = x / (n * K0)
    d2 = d * d
    d3 = d2 * d
    d4 = d3 * d
    d5 = d4 * d
    d6 = d5 * d

    latitude = (
        p_rad
        - (p_tan / r)
        * (d2 / 2 - d4 / 24 * (5 + 3 * p_tan2 + 10 * c - 4 * c2 - 9 * E_P2))
        + d6 / 720 * (61 + 90 * p_tan2 + 298 * c + 45 * p_tan4 - 252 * E_P2 - 3 * c2)
    )

    longitude = (
        d
        - d3 / 6 * (1 + 2 * p_tan2 + c)
        + d5 / 120 * (5 - 2 * c + 28 * p_tan2 - 3 * c2 + 8 * E_P2 + 24 * p_tan4)
    ) / p_cos

    return (
        numpy.degrees(latitude),
        numpy.degrees(longitude) + zone_number_to_central_longitude(zone_number),
    )


def from_latlon(
    latitude: numpy.array,
    longitude: numpy.array,
    force_zone_number: int = None,
    force_zone_letter: str = None,
) -> (numpy.array, numpy.array, numpy.array, numpy.array):
    """
    This function convert Latitude and Longitude to UTM coordinate

    :param latitude: float or array. Latitude between 80 deg S and 84 deg N, e.g. (-80.0 to 84.0)
    :param longitude: float or array: Longitude between 180 deg W and 180 deg E, e.g. (-180.0 to 180.0).
    :param force_zone number: int: Zone Number is represented with global map numbers of an UTM Zone
        Numbers Map. You may force conversion including one UTM Zone Number.
        More information see http://www.jaworski.ca/utmzones.htm
    :return: eastings, northings, zone_number, zone_letter
    """
    import rsgislib.tools.utils

    if not rsgislib.tools.utils.in_bounds(latitude, -80.0, 84.0):
        raise rsgislib.RSGISPyException(
            "latitude out of range (must be between 80 deg S and 84 deg N)"
        )
    if not rsgislib.tools.utils.in_bounds(longitude, -180.0, 180.0):
        raise rsgislib.RSGISPyException(
            "longitude out of range (must be between 180 deg W and 180 deg E)"
        )
    if force_zone_number is not None:
        check_valid_zone(force_zone_number, force_zone_letter)

    lat_rad = numpy.radians(latitude)
    lat_sin = numpy.sin(lat_rad)
    lat_cos = numpy.cos(lat_rad)

    lat_tan = lat_sin / lat_cos
    lat_tan2 = lat_tan * lat_tan
    lat_tan4 = lat_tan2 * lat_tan2

    if force_zone_number is None:
        zone_number = latlon_to_zone_number(latitude, longitude)
    else:
        zone_number = force_zone_number

    if force_zone_letter is None:
        zone_letter = latitude_to_zone_letter(latitude)
    else:
        zone_letter = force_zone_letter

    lon_rad = numpy.radians(longitude)
    central_lon = zone_number_to_central_longitude(zone_number)
    central_lon_rad = numpy.radians(central_lon)

    n = R / numpy.sqrt(1 - E * lat_sin**2)
    c = E_P2 * lat_cos**2

    a = lat_cos * (lon_rad - central_lon_rad)
    a2 = a * a
    a3 = a2 * a
    a4 = a3 * a
    a5 = a4 * a
    a6 = a5 * a

    m = R * (
        M1 * lat_rad
        - M2 * numpy.sin(2 * lat_rad)
        + M3 * numpy.sin(4 * lat_rad)
        - M4 * numpy.sin(6 * lat_rad)
    )

    easting = (
        K0
        * n
        * (
            a
            + a3 / 6 * (1 - lat_tan2 + c)
            + a5 / 120 * (5 - 18 * lat_tan2 + lat_tan4 + 72 * c - 58 * E_P2)
        )
        + 500000
    )

    northing = K0 * (
        m
        + n
        * lat_tan
        * (
            a2 / 2
            + a4 / 24 * (5 - lat_tan2 + 9 * c + 4 * c**2)
            + a6 / 720 * (61 - 58 * lat_tan2 + lat_tan4 + 600 * c - 330 * E_P2)
        )
    )

    if rsgislib.tools.utils.mixed_signs(latitude):
        raise ValueError("latitudes must all have the same sign")
    elif rsgislib.tools.utils.negative(latitude):
        northing += 10000000

    return easting, northing, zone_number, zone_letter


def latitude_to_zone_letter(latitude):
    """
    Find the zone letter for the latitude. If the input is a numpy array, just use the first element
    user responsibility to make sure that all points are in one zone

    :param latitude:
    :return: string if found otherwise None.
    """
    if isinstance(latitude, numpy.ndarray):
        latitude = latitude.flat[0]

    if -80 <= latitude <= 84:
        return ZONE_LETTERS[int(latitude + 80) >> 3]
    else:
        return None


def latlon_to_zone_number(latitude, longitude):
    """
    Find the UTM zone number for a give latitude and longitude. If the input is a numpy array, just use the
    first element user responsibility to make sure that all points are in one zone

    :param latitude: float
    :param longitude: float

    :return: int
    """
    if isinstance(latitude, numpy.ndarray):
        latitude = latitude.flat[0]
    if isinstance(longitude, numpy.ndarray):
        longitude = longitude.flat[0]

    if 56 <= latitude < 64 and 3 <= longitude < 12:
        return 32

    if 72 <= latitude <= 84 and longitude >= 0:
        if longitude < 9:
            return 31
        elif longitude < 21:
            return 33
        elif longitude < 33:
            return 35
        elif longitude < 42:
            return 37

    return int((longitude + 180) / 6) + 1


def latlon_arr_to_utm_zone_number(
    latitude: numpy.array, longitude: numpy.array
) -> numpy.array:
    """
    Find the UTM zone number for a give latitude and longitude. UTM zone will be
    returned for all the lat/longs within the input arrays, which must be of the same
    length. Function will also work with a single value, at which point a single
    int will be returned.

    :param latitude: numpy array of floats
    :param longitude: numpy array of floats

    :return: int or array of ints.
    """
    # utm_zones = numpy.zeros_like(latitude, dtype=numpy.dtype(int))
    utm_zones = ((longitude + 180) / 6) + 1
    utm_zones = numpy.rint(utm_zones).astype(int)

    utm_zones[
        (56 <= latitude) & (latitude < 64) & (3 <= longitude) & (longitude < 12)
    ] = 32
    utm_zones[
        (72 <= latitude) & (latitude <= 84) & (longitude >= 0) & (longitude < 9)
    ] = 31
    utm_zones[
        (72 <= latitude) & (latitude <= 84) & (longitude >= 0) & (longitude < 21)
    ] = 33
    utm_zones[
        (72 <= latitude) & (latitude <= 84) & (longitude >= 0) & (longitude < 33)
    ] = 35
    utm_zones[
        (72 <= latitude) & (latitude <= 84) & (longitude >= 0) & (longitude < 42)
    ] = 37

    return utm_zones


def latlon_to_mode_utm_zone_number(
    latitude: numpy.array, longitude: numpy.array
) -> int:
    """
    Find the mode UTM zone for a list of lat/lon values.

    :param latitude: numpy array of floats
    :param longitude: numpy array of floats
    :return: int (mode UTM zone)

    """
    import scipy.stats

    utm_zones = latlon_arr_to_utm_zone_number(latitude, longitude)
    mode, count = scipy.stats.mode(utm_zones)
    return mode[0]


def zone_number_to_central_longitude(zone_number: int) -> int:
    """
    Find the central longitude for the given zone.

    :param zone_number: int
    :return: float
    """
    return (zone_number - 1) * 6 - 180 + 3



#!/usr/bin/env python
############################################################################
#  solarangles.py
#
#  Copyright 2016 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose: Provide a set of utilities which provide further functionality
#          specifically for caclulating solar angles.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 23/11/2016
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

from osgeo import osr
import numpy


def get_solar_irr_convention_solar_azimuth_from_usgs(solarAz):
    """
    IN: USGS Convertion::

                  N (0)
                  |
         W (-90)-----E (90)
                  |
           (-180) S (180)

    OUT: Solar Irradiance Convertion::

                  N (0)
                  |
         W (270)-----E (90)
                  |
                  S (180)
    """
    solarAzOut = solarAz
    if solarAz < 0:
        solarAzOut = 360 + solarAz
    return solarAzOut


def get_solar_irr_convention_solar_azimuth_from_trad(solarAz):
    """
    IN: Traditional Convertion::

           (-180) N (180)
                  |
         W (-90)-----E (90)
                  |
                  S (0)

    OUT: Solar Irradiance Convertion::

                  N (0)
                  |
         W (270)-----E (90)
                  |
                  S (180)
    """
    solarAzOut = 0.0
    if solarAz > 0:
        solarAzOut = 180 - solarAz
    elif solarAz < 0:
        solarAzOut = 180 + ((-1) * solarAz)
    return solarAzOut


def calc_solar_azimuth_zenith(inputImg, inImgDateTime, outputImg, gdalformat):
    """
    Function which calculate a solar azimuth (band 1) and zenith (band 2) image.

    :param inputImg: input image file (can be any image with a spatial header)
    :param inImgDateTime: a datatime object for the data/time of the acquasition
    :param outputImg: output image file and path
    :param gdalformat: output file format (e.g., KEA)
    """
    import Pysolar
    from rios import applier
    from rios import cuiprogress
    import rsgislib

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    infiles = applier.FilenameAssociations()
    infiles.image1 = inputImg
    outfiles = applier.FilenameAssociations()
    outfiles.outimage = outputImg
    otherargs = applier.OtherInputs()
    otherargs.dateTime = inImgDateTime
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat

    wgs84latlonProj = osr.SpatialReference()
    wgs84latlonProj.ImportFromEPSG(4326)
    otherargs.wgs84latlonProj = wgs84latlonProj

    def _calcSolarAzimuthZenith(info, inputs, outputs, otherargs):
        """
        Internal functions used within calc_solar_azimuth_zenith() - don't call independently.

        """
        xBlock, yBlock = info.getBlockCoordArrays()

        inProj = osr.SpatialReference()
        inProj.ImportFromWkt(info.getProjection())

        transform = osr.CoordinateTransformation(inProj, otherargs.wgs84latlonProj)
        xBlockF = xBlock.flatten()
        yBlockF = yBlock.flatten()

        pts = numpy.zeros((xBlockF.shape[0], 2), dtype=numpy.float32)
        pts[..., 0] = xBlockF
        pts[..., 1] = yBlockF
        outPts = numpy.array(transform.TransformPoints(pts))

        outAz = numpy.zeros_like(xBlockF, dtype=numpy.float32)
        outZen = numpy.zeros_like(xBlockF, dtype=numpy.float32)

        for i in range(outPts.shape[0]):
            outAz[i] = 90 - Pysolar.solar.GetAltitude(
                outPts[i, 1], outPts[i, 0], otherargs.dateTime
            )
            outZen[i] = (
                (-1)
                * Pysolar.solar.GetAzimuth(
                    outPts[i, 1], outPts[i, 0], otherargs.dateTime
                )
            ) - 180

        outAz = numpy.reshape(outAz, xBlock.shape)
        outZen = numpy.reshape(outZen, xBlock.shape)

        # print("Block End")
        outputs.outimage = numpy.stack((outAz, outZen))

    # Apply the multiply function.
    applier.apply(
        _calcSolarAzimuthZenith, infiles, outfiles, otherargs, controls=aControls
    )



#! /usr/bin/env python
############################################################################
#  tiledclump.py
#
#  Copyright 2016 RSGISLib.
#
#  RSGISLib: 'The Remote Sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide functionality to clump an image using image tiles.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 21/3/2016
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

import glob
import os.path
import os
import shutil
from multiprocessing import Pool
import multiprocessing

import rsgislib
from rsgislib import segmentation
from rsgislib import imageutils

################################ Clumping Functions ################################


def perform_clumping_single_thread(
    input_img, clumps_img, tmp_dir="tmp", width=2000, height=2000, gdalformat="KEA"
):
    """
    Clump the input image using a tiled processing chain allowing large images to
    be clumped more quickly.

    :param input_img: the input image to be clumped.
    :param clumps_img: the output clumped image.
    :param tmp_dir: the temporary directory where intermediate files will be
                    written (default is 'tmp'). Directory will be created and
                    deleted if does not exist.
    :param width: int for width of the image tiles used for
                  processing (Default = 2000).
    :param height: int for height of the image tiles used for
                   processing (Default = 2000).
    :param gdalformat: string with the GDAL image format for the output image
                       (Default = KEA). NOTE. KEA is used as intermediate format
                       internally and therefore needs to be available.

    """
    import rsgislib.tools.utils
    import rsgislib.tools.filetools

    createdTmp = False
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
        createdTmp = True

    uidStr = rsgislib.tools.utils.uid_generator()
    dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
    baseName = os.path.splitext(os.path.basename(input_img))[0] + "_" + uidStr
    imgTilesDIR = os.path.join(tmp_dir, "imgtiles_" + uidStr)
    tilesClumpsDIR = os.path.join(tmp_dir, "imgclumpstiles_" + uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)
    initMergedClumps = os.path.join(tmp_dir, "MergedInitClumps_" + uidStr + ".kea")
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    if not os.path.exists(tilesClumpsDIR):
        os.makedirs(tilesClumpsDIR)

    imageutils.create_tiles(
        input_img,
        tilesImgBase,
        int(width),
        int(height),
        0,
        False,
        "KEA",
        dataType,
        "kea",
    )
    imageTiles = glob.glob(tilesImgBase + "*")

    for tile in imageTiles:
        tilBaseName = os.path.splitext(os.path.basename(tile))[0]
        clumpedTile = os.path.join(tilesClumpsDIR, tilBaseName + "_clumps.kea")
        segmentation.clump(tile, clumpedTile, "KEA", True, 0, True)

    clumpTiles = glob.glob(os.path.join(tilesClumpsDIR, "*_clumps.kea"))
    print("Create Blank Image")
    imageutils.create_copy_img(
        input_img, initMergedClumps, 1, 0, "KEA", rsgislib.TYPE_32UINT
    )
    print("Merge Tiles into Blank Image")
    segmentation.merge_clump_images(clumpTiles, initMergedClumps, True)
    print("Merge Tile Boundaries")
    segmentation.merge_equiv_clumps(
        initMergedClumps, clumps_img, gdalformat, ["PixelVal"]
    )

    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesClumpsDIR)
    rsgislib.tools.filetools.delete_file_silent(initMergedClumps)
    if createdTmp:
        shutil.rmtree(tmp_dir)


def _clump_img_func(imgs):
    """
    Clump an image with values provides as an array for use within a multiprocessing Pool
    """
    segmentation.clump(imgs[0], imgs[1], "KEA", True, 0, True)


def perform_clumping_multi_process(
    input_img,
    clumps_img,
    tmp_dir="tmp",
    width=2000,
    height=2000,
    gdalformat="KEA",
    n_cores=-1,
):
    """
    Clump the input image using a tiled processing chain allowing large images
    to be clumped more quickly.

    :param input_img: the input image to be clumped.
    :param clumps_img: the output clumped image.
    :param tmp_dir: the temporary directory where intermediate files will be
                    written (default is 'tmp'). Directory will be created and
                    deleted if does not exist.
    :param width: int for width of the image tiles used for
                  processing (Default = 2000).
    :param height: int for height of the image tiles used for
                   processing (Default = 2000).
    :param gdalformat: string with the GDAL image format for the output image
                       (Default = KEA). NOTE. KEA is used as intermediate format
                       internally and therefore needs to be available.
    :param n_cores: is an int specifying the number of cores to be used for
                    clumping processing.

    """
    import rsgislib.tools.utils
    import rsgislib.tools.filetools

    createdTmp = False
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
        createdTmp = True

    if n_cores <= 0:
        n_cores = multiprocessing.cpu_count()

    uidStr = rsgislib.tools.utils.uid_generator()
    dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
    baseName = os.path.splitext(os.path.basename(input_img))[0] + "_" + uidStr
    imgTilesDIR = os.path.join(tmp_dir, "imgtiles_" + uidStr)
    tilesClumpsDIR = os.path.join(tmp_dir, "imgclumpstiles_" + uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)
    initMergedClumps = os.path.join(tmp_dir, "MergedInitClumps_" + uidStr + ".kea")
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    if not os.path.exists(tilesClumpsDIR):
        os.makedirs(tilesClumpsDIR)

    imageutils.create_tiles(
        input_img,
        tilesImgBase,
        int(width),
        int(height),
        0,
        False,
        "KEA",
        dataType,
        "kea",
    )
    imageTiles = glob.glob(tilesImgBase + "*")

    clumpImgsVals = []
    for tile in imageTiles:
        tilBaseName = os.path.splitext(os.path.basename(tile))[0]
        clumpedTile = os.path.join(tilesClumpsDIR, tilBaseName + "_clumps.kea")
        clumpImgsVals.append([tile, clumpedTile])

    with Pool(n_cores) as p:
        p.map(_clump_img_func, clumpImgsVals)

    clumpTiles = glob.glob(os.path.join(tilesClumpsDIR, "*_clumps.kea"))
    print("Create Blank Image")
    imageutils.create_copy_img(
        input_img, initMergedClumps, 1, 0, "KEA", rsgislib.TYPE_32UINT
    )
    print("Merge Tiles into Blank Image")
    segmentation.merge_clump_images(clumpTiles, initMergedClumps, True)
    print("Merge Tile Boundaries")
    segmentation.merge_equiv_clumps(
        initMergedClumps, clumps_img, gdalformat, ["PixelVal"]
    )

    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesClumpsDIR)
    rsgislib.tools.filetools.delete_file_silent(initMergedClumps)
    if createdTmp:
        shutil.rmtree(tmp_dir)


################################################################################################


################################ Union Clumping Functions ################################


def perform_union_clumping_single_thread(
    input_img,
    in_ref_img,
    clumps_img,
    tmp_dir="tmp",
    width=2000,
    height=2000,
    gdalformat="KEA",
):
    """
    Clump and union with the reference image the input image using a tiled processing
    chain allowing large images to be clumped more quickly.

    :param input_img: the input image to be clumped.
    :param in_ref_img: the reference image which the union is undertaken with
                       (typically an existing classification)
    :param clumps_img: the output clumped image.
    :param tmp_dir: the temporary directory where intermediate files will be written
                    (default is 'tmp'). Directory will be created and deleted if
                    does not exist.
    :param width: int for width of the image tiles used for
                  processing (Default = 2000).
    :param height: int for height of the image tiles used for
                   processing (Default = 2000).
    :param gdalformat: string with the GDAL image format for the output image
                       (Default = KEA). NOTE. KEA is used as intermediate format
                       internally and therefore needs to be available.

    """
    import rsgislib.tools.utils
    import rsgislib.tools.filetools

    createdTmp = False
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
        createdTmp = True

    uidStr = rsgislib.tools.utils.uid_generator()
    dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
    baseName = os.path.splitext(os.path.basename(input_img))[0] + "_" + uidStr
    imgTilesDIR = os.path.join(tmp_dir, "imgtiles_" + uidStr)
    tilesClumpsDIR = os.path.join(tmp_dir, "imgclumpstiles_" + uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)
    initMergedClumps = os.path.join(tmp_dir, "MergedInitClumps_" + uidStr + ".kea")
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    if not os.path.exists(tilesClumpsDIR):
        os.makedirs(tilesClumpsDIR)

    imageutils.create_tiles(
        input_img,
        tilesImgBase,
        int(width),
        int(height),
        0,
        False,
        "KEA",
        dataType,
        "kea",
    )
    imageTiles = glob.glob(tilesImgBase + "*")

    for tile in imageTiles:
        tilBaseName = os.path.splitext(os.path.basename(tile))[0]
        clumpedTile = os.path.join(tilesClumpsDIR, tilBaseName + "_clumps.kea")
        segmentation.union_of_clumps([tile, in_ref_img], clumpedTile, "KEA", 0, True)

    clumpTiles = glob.glob(os.path.join(tilesClumpsDIR, "*_clumps.kea"))
    print("Create Blank Image")
    imageutils.create_copy_img(
        input_img, initMergedClumps, 1, 0, "KEA", rsgislib.TYPE_32UINT
    )
    print("Merge Tiles into Blank Image")
    segmentation.merge_clump_images(clumpTiles, initMergedClumps, True)
    print("Merge Tile Boundaries")
    segmentation.merge_equiv_clumps(
        initMergedClumps, clumps_img, gdalformat, ["ClumpVal_1", "ClumpVal_2"]
    )

    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesClumpsDIR)
    rsgislib.tools.filetools.delete_file_silent(initMergedClumps)
    if createdTmp:
        shutil.rmtree(tmp_dir)


def _union_clump_img_func(imgs):
    """
    Union Clump an image with values provides as an array for use
    within a multiprocessing Pool
    """
    segmentation.union_of_clumps([imgs[0], imgs[1]], imgs[2], "KEA", 0, True)


def perform_union_clumping_multi_process(
    input_img,
    in_ref_img,
    clumps_img,
    tmp_dir="tmp",
    width=2000,
    height=2000,
    gdalformat="KEA",
    n_cores=-1,
):
    """
    Clump and union with the reference image the input image using a tiled
    processing chain allowing large images to be clumped more quickly.

    :param input_img: the input image to be clumped.
    :param in_ref_img: the reference image which the union is undertaken
                       with (typically an existing classification)
    :param clumps_img: the output clumped image.
    :param tmp_dir: the temporary directory where intermediate files will be
                    written (default is 'tmp'). Directory will be created and
                    deleted if does not exist.
    :param width: int for width of the image tiles used for
                  processing (Default = 2000).
    :param height: int for height of the image tiles used for
                   processing (Default = 2000).
    :param gdalformat: string with the GDAL image format for the output image
                       (Default = KEA). NOTE. KEA is used as intermediate format
                       internally and therefore needs to be available.
    :param n_cores: is an int specifying the number of cores to be used for
                    clumping processing.

    """
    import rsgislib.tools.utils
    import rsgislib.tools.filetools

    createdTmp = False
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
        createdTmp = True

    if n_cores <= 0:
        n_cores = multiprocessing.cpu_count()

    uidStr = rsgislib.tools.utils.uid_generator()
    dataType = rsgislib.imageutils.get_rsgislib_datatype_from_img(input_img)
    baseName = os.path.splitext(os.path.basename(input_img))[0] + "_" + uidStr
    imgTilesDIR = os.path.join(tmp_dir, "imgtiles_" + uidStr)
    tilesClumpsDIR = os.path.join(tmp_dir, "imgclumpstiles_" + uidStr)
    tilesImgBase = os.path.join(imgTilesDIR, baseName)
    initMergedClumps = os.path.join(tmp_dir, "MergedInitClumps_" + uidStr + ".kea")
    if not os.path.exists(imgTilesDIR):
        os.makedirs(imgTilesDIR)
    if not os.path.exists(tilesClumpsDIR):
        os.makedirs(tilesClumpsDIR)

    imageutils.create_tiles(
        input_img,
        tilesImgBase,
        int(width),
        int(height),
        0,
        False,
        "KEA",
        dataType,
        "kea",
    )
    imageTiles = glob.glob(tilesImgBase + "*")

    clumpImgsVals = []
    for tile in imageTiles:
        tilBaseName = os.path.splitext(os.path.basename(tile))[0]
        clumpedTile = os.path.join(tilesClumpsDIR, tilBaseName + "_clumps.kea")
        clumpImgsVals.append([tile, in_ref_img, clumpedTile])

    with Pool(n_cores) as p:
        p.map(_union_clump_img_func, clumpImgsVals)

    clumpTiles = glob.glob(os.path.join(tilesClumpsDIR, "*_clumps.kea"))
    print("Create Blank Image")
    imageutils.create_copy_img(
        input_img, initMergedClumps, 1, 0, "KEA", rsgislib.TYPE_32UINT
    )
    print("Merge Tiles into Blank Image")
    segmentation.merge_clump_images(clumpTiles, initMergedClumps, True)
    print("Merge Tile Boundaries")
    segmentation.merge_equiv_clumps(
        initMergedClumps, clumps_img, gdalformat, ["ClumpVal_1", "ClumpVal_2"]
    )

    shutil.rmtree(imgTilesDIR)
    shutil.rmtree(tilesClumpsDIR)
    rsgislib.tools.filetools.delete_file_silent(initMergedClumps)
    if createdTmp:
        shutil.rmtree(tmp_dir)



#!/usr/bin/env python

#############################################
# rsgisbuildimglut.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose:  Build a (vector) look up table (LUT) for a directory of 
#           images.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 01/02/2019
# Version: 1.0
# 
#############################################

import glob
import os.path
import argparse

import rsgislib.imageutils.imagelut

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--imgsearch", type=str, required=True, help="""Input search path which includes wildcard (i.e., *) 
                                                                        for python glob function to find the images to build
                                                                        the LUT. Note. This input will need to be provided in \"quotes\".""")
    parser.add_argument("-o", "--output", type=str, required=True, help="Output vector file representing the LUT.")
    parser.add_argument("--veclyr", type=str, required=False, help="Alternative layer name (if supported) for the LUT vector file.")
    parser.add_argument("--vecformat", type=str, required=False, default="GPKG", help="OGR vector format for the LUT file.")
    parser.add_argument("--wgs84", action='store_true', default=False, help="Reproject footprints to EPSG:4326 (WGS84).")
    parser.add_argument("--ignore_noproj", action='store_true', default=False, help="Silently ignored images which do not have a defined projection.")
    parser.add_argument("--overwrite", action='store_true', default=False, help="Overwrite the output vector file, if not specified then layer will be added to exisitng file.")
    
    args = parser.parse_args()
    
    if args.veclyr is None:
        vec_lyr_name = os.path.splitext(os.path.basename(args.output))[0]
    else:
        vec_lyr_name = args.veclyr
    
    img_list = glob.glob(args.imgsearch)
    rsgislib.imageutils.imagelut.create_img_extent_lut(img_list, args.output, vec_lyr_name, args.vecformat, ignore_none_imgs=args.ignore_noproj, out_proj_wgs84=args.wgs84, overwrite_lut_file=args.overwrite)
    print("Finished Creating LUT.")



#!/usr/bin/env python
############################################################################
#  sensorlvl2data.py
#
#  Copyright 2021 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose: Provide a set of utilities for using downloaded level 2 data
#          from data providers in an easy manner.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 10/10/2021
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
############################################################################

import rsgislib


def create_stacked_ls8_cl2_lv2_img(
    input_file,
    out_dir,
    tmp_dir,
    scale_factor=10000,
    gdalformat="KEA",
    delete_inter_data=True,
):
    """
    A function which extracts the USGS collection-2 level-2 data and creates a single
    multi-band image file in the output directory. Note the output data type is
    32bit integer and the no data value is -9999.

    :param input_file: file name and path to the input TAR file.
    :param out_dir: output directory where the output file will be written.
    :param tmp_dir: a temporary directory where intermediate files will be written.
    :param scale_factor: the scale factor applied to the output reflectance values.
                         default is 10,000
    :param gdalformat: the output file format.
    :param delete_inter_data: delete any intermediate files created is True (Default)

    .. code:: python

        import rsgislib.imagecalibration.sensorlvl2data
        rsgislib.imagecalibration.sensorlvl2data.create_stacked_ls8_cl2_lv2_img("LC08_L2SP_135046_20210319_20210328_02_T1.tar", "outputs", "tmp", gdalformat="GTIFF")

    """

    import os
    import rsgislib.tools.filetools
    import rsgislib.tools.sensors
    import rsgislib.tools.utils
    import rsgislib.imagecalc
    import rsgislib.imageutils

    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)

    if gdalformat == "GTIFF":
        rsgislib.imageutils.set_env_vars_lzw_gtiff_outs()

    raw_dir = rsgislib.tools.filetools.untar_file(input_file, tmp_dir)

    mtl_header = rsgislib.tools.filetools.find_file_none(raw_dir, "*MTL.txt")

    if mtl_header is None:
        raise rsgislib.RSGISPyException("Could not find the MTL header")

    ls8_head_info = rsgislib.tools.sensors.read_landsat_mtl_to_dict(mtl_header)

    ls_prod_id = rsgislib.tools.utils.dict_struct_get_str_value(
        ls8_head_info, ["PRODUCT_CONTENTS", "LANDSAT_PRODUCT_ID"]
    ).lower()

    b1_img = rsgislib.tools.utils.dict_struct_get_str_value(
        ls8_head_info, ["PRODUCT_CONTENTS", "FILE_NAME_BAND_1"]
    )
    b2_img = rsgislib.tools.utils.dict_struct_get_str_value(
        ls8_head_info, ["PRODUCT_CONTENTS", "FILE_NAME_BAND_2"]
    )
    b3_img = rsgislib.tools.utils.dict_struct_get_str_value(
        ls8_head_info, ["PRODUCT_CONTENTS", "FILE_NAME_BAND_3"]
    )
    b4_img = rsgislib.tools.utils.dict_struct_get_str_value(
        ls8_head_info, ["PRODUCT_CONTENTS", "FILE_NAME_BAND_4"]
    )
    b5_img = rsgislib.tools.utils.dict_struct_get_str_value(
        ls8_head_info, ["PRODUCT_CONTENTS", "FILE_NAME_BAND_5"]
    )
    b6_img = rsgislib.tools.utils.dict_struct_get_str_value(
        ls8_head_info, ["PRODUCT_CONTENTS", "FILE_NAME_BAND_6"]
    )
    b7_img = rsgislib.tools.utils.dict_struct_get_str_value(
        ls8_head_info, ["PRODUCT_CONTENTS", "FILE_NAME_BAND_7"]
    )

    b1_multi = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_MULT_BAND_1"],
    )
    b2_multi = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_MULT_BAND_2"],
    )
    b3_multi = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_MULT_BAND_3"],
    )
    b4_multi = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_MULT_BAND_4"],
    )
    b5_multi = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_MULT_BAND_5"],
    )
    b6_multi = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_MULT_BAND_6"],
    )
    b7_multi = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_MULT_BAND_7"],
    )

    b1_add = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_ADD_BAND_1"],
    )
    b2_add = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_ADD_BAND_2"],
    )
    b3_add = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_ADD_BAND_3"],
    )
    b4_add = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_ADD_BAND_4"],
    )
    b5_add = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_ADD_BAND_5"],
    )
    b6_add = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_ADD_BAND_6"],
    )
    b7_add = rsgislib.tools.utils.dict_struct_get_numeric_value(
        ls8_head_info,
        ["LEVEL2_SURFACE_REFLECTANCE_PARAMETERS", "REFLECTANCE_ADD_BAND_7"],
    )

    band_info = []
    band_info.append(
        {"img": os.path.join(raw_dir, b1_img), "multi": b1_multi, "add": b1_add}
    )
    band_info.append(
        {"img": os.path.join(raw_dir, b2_img), "multi": b2_multi, "add": b2_add}
    )
    band_info.append(
        {"img": os.path.join(raw_dir, b3_img), "multi": b3_multi, "add": b3_add}
    )
    band_info.append(
        {"img": os.path.join(raw_dir, b4_img), "multi": b4_multi, "add": b4_add}
    )
    band_info.append(
        {"img": os.path.join(raw_dir, b5_img), "multi": b5_multi, "add": b5_add}
    )
    band_info.append(
        {"img": os.path.join(raw_dir, b6_img), "multi": b6_multi, "add": b6_add}
    )
    band_info.append(
        {"img": os.path.join(raw_dir, b7_img), "multi": b7_multi, "add": b7_add}
    )

    sref_bands = []
    for i, band in enumerate(band_info):
        out_img_band = os.path.join(
            tmp_dir, "{}_b{}_sref.kea".format(ls_prod_id, i + 1)
        )
        rsgislib.imagecalc.image_math(
            band["img"],
            out_img_band,
            "b1==0?-9999:((b1*{})+{})*{}".format(
                band["multi"], band["add"], scale_factor
            ),
            "KEA",
            rsgislib.TYPE_32INT,
        )
        sref_bands.append(out_img_band)

    band_names = ["coastal", "blue", "green", "red", "nir", "swir1", "swir2"]
    out_img_ext = rsgislib.imageutils.get_file_img_extension(gdalformat)
    output_img = os.path.join(out_dir, "{}_sref.{}".format(ls_prod_id, out_img_ext))
    rsgislib.imageutils.stack_img_bands(
        sref_bands, band_names, output_img, None, -9999, gdalformat, rsgislib.TYPE_32INT
    )
    rsgislib.imageutils.pop_img_stats(
        output_img, use_no_data=True, no_data_val=-9999, calc_pyramids=True
    )

    if delete_inter_data:
        import shutil

        shutil.rmtree(raw_dir)
        for img_band in sref_bands:
            rsgislib.imageutils.delete_gdal_layer(img_band)



import os

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
SEGMENT_DATA_DIR = os.path.join(DATA_DIR, "segmentation")
IMGCALC_DATA_DIR = os.path.join(DATA_DIR, "imagecalc")


def test_perform_clumping_single_thread(tmp_path):
    import rsgislib.segmentation.tiledclump

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    clumps_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.segmentation.tiledclump.perform_clumping_single_thread(
        input_img, clumps_img, tmp_dir=tmp_path, width=500, height=500, gdalformat="KEA"
    )
    assert os.path.exists(clumps_img)


def test_perform_clumping_multi_process(tmp_path):
    import rsgislib.segmentation.tiledclump

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    clumps_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.segmentation.tiledclump.perform_clumping_multi_process(
        input_img,
        clumps_img,
        tmp_dir=tmp_path,
        width=500,
        height=500,
        gdalformat="KEA",
        n_cores=1,
    )
    assert os.path.exists(clumps_img)


def test_perform_union_clumping_single_thread(tmp_path):
    import rsgislib.segmentation.tiledclump

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    clumps_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.segmentation.tiledclump.perform_union_clumping_single_thread(
        input_img,
        input_img,
        clumps_img,
        tmp_dir=tmp_path,
        width=500,
        height=500,
        gdalformat="KEA",
    )
    assert os.path.exists(clumps_img)


def test_perform_union_clumping_multi_process(tmp_path):
    import rsgislib.segmentation.tiledclump

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    clumps_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.segmentation.tiledclump.perform_union_clumping_multi_process(
        input_img,
        input_img,
        clumps_img,
        tmp_dir=tmp_path,
        width=500,
        height=500,
        gdalformat="KEA",
        n_cores=1,
    )
    assert os.path.exists(clumps_img)



import os
import pytest

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


SKIMAGE_NOT_AVAIL = False
try:
    import skimage
except ImportError:
    SKIMAGE_NOT_AVAIL = True


@pytest.mark.skipif(SKIMAGE_NOT_AVAIL, reason="scikit-image dependency not available")
def test_perform_least_cost_path_calc(tmp_path):
    import rsgislib.imagecalc.leastcostpath

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    start_coord = (257938, 280795)
    stop_coord = (260201, 280445)
    rsgislib.imagecalc.leastcostpath.perform_least_cost_path_calc(
        input_img,
        output_img,
        start_coord,
        stop_coord,
        gdalformat="KEA",
        cost_img_band=1,
    )
    assert os.path.exists(output_img)
